# AUTOGENERATED! DO NOT EDIT! File to edit: ../notebooks/08_geotools.ipynb.

# %% auto 0
__all__ = ['calculate_image_azimuth', 'get_north_shifted_point', 'calculate_image_north_azimuth', 'get_sun_angles', 'debug_srs',
           'shift_to_center', 'pixel_to_meter', 'Point', 'Window', 'ImgData']

# %% ../notebooks/08_geotools.ipynb 3
import matplotlib.cm as cm
import matplotlib.pyplot as plt
import numpy as np
from matplotlib.pyplot import figure, show
from mpl_toolkits.axes_grid1.anchored_artists import AnchoredSizeBar
try:
    from osgeo import gdal, osr
except ImportError:
    print("GDAL not installed. The `geotools` module requires it.")

from .exceptions import ProjectionNotSetError, SomethingNotSetError

gdal.UseExceptions()

# %% ../notebooks/08_geotools.ipynb 4
def calculate_image_azimuth(origPoint, newPoint, zero="right"):
    """Calculate azimuth angle between 2 image points.

    Beware that this function calculates trigonometric angles.
    If the points are from an image that has (0, 0) in the upper left, this means
    that the angles increase clockwise.
    That is why, for example, for an HiRISE image, the return of this function
    matches the angle rotation definition for HiRISE data.

    Parameters
    ==========
    origPoint, newPoint: <mars.Point> objects
    zero: <string>
        right' or 'top' to indicate where zero is.

    Returns
    =======
    azimuth: <float>
        Azimuth angle
    """
    deltaSample = newPoint.sample - origPoint.sample
    deltaLine = newPoint.line - origPoint.line

    azimuth = np.degrees(np.arctan2(deltaLine, deltaSample))

    if azimuth < 0.0:
        azimuth += 360.0
    if azimuth > 360.0:
        azimuth -= 360.0
    if zero == "top":
        azimuth += 90.0
        if azimuth > 360.0:
            azimuth -= 360.0

    return azimuth


def get_north_shifted_point(img, offset=0.001):
    "Increasing the latitude is a sure way of getting north."
    newPoint = Point(lon=img.center.lon, lat=img.center.lat + offset)
    newPoint.lonlat_to_pixel(img.geotrans, img.projection)
    return newPoint


def calculate_image_north_azimuth(img, zero="right"):
    newPoint = get_north_shifted_point(img)
    return img.center.calculate_azimuth(newPoint, zero=zero)


def get_sun_angles(spicer, img):
    """Calculate solar azimuth and incidence.

    By requiring a spicer object for this function, it becomes independent from the
    solar system object where the calculations are made.

    Parameters
    ----------
    spicer : spicer.Spicer
        needs to be setup for time, but spoint is set from img.center in here.
    img : geotools.ImgData
        The image data of which the center point serves as the start point.

    Returns
    -------
    tuple(float, float)
        Solar azimuth, incidence [degrees]
    """
    p = img.center
    spicer.set_spoint_by(lat=p.lat, lon=p.lon)
    p2lon, p2lat = spicer.point_towards_sun(pixel_res=1)
    p2 = Point.copy_geodata(p, lat=p2lat.value, lon=p2lon.value)
    p2.lonlat_to_pixel()
    # remember, this calculates CCW angles, but if image is "origin='upper'",
    # it becomes CW
    sun_az = p.calculate_azimuth(p2)
    sun_inc = spicer.illum_angles.dsolar
    return sun_az, sun_inc.value


def debug_srs(projection):
    """Correct wrong scale_factor in PolarStereographic data.

    Some PolarStereographic data have a 0 as a scale_factor in the
    projection (mostly MOLA), which is being corrected here.
    TODO: Check for being PolarStereographic before doing this!
    """
    srs = osr.SpatialReference(projection)
    if int(srs.GetProjParm("scale_factor")) == 0:
        srs.SetProjParm("scale_factor", 1)
    return srs


def shift_to_center(x, y, geotransform):
    # if i'd shift, the centerpoint does not show center coordinates
    # so that seems wrong. am i overlooking something?
    x += geotransform[1] / 2.0
    y += geotransform[5] / 2.0
    return x, y


def pixel_to_meter(sample, line, geotransform, shift=False):
    """provide point in map projection coordinates.

    Parameters
    ==========
    sample, line: <integer> Sample and line of an image in pixel coords
    Geotransform in format as given by GDAL datasets.GetGeoTransform()

    Returns
    =======
    tuple (x,y) coordinates in the projection of the dataset

    TODO: Shift had a problem. Write test to confirm what and fix.
    FIXME above
    """
    x, y = gdal.ApplyGeoTransform(geotransform, sample, line)
    if shift:
        x, y = shift_to_center(x, y, geotransform)
    return (x, y)

# %% ../notebooks/08_geotools.ipynb 5
class Point:
    """Point class to manage pixel and map points and their transformations.

    Parameters
    ==========
    Either:
    sample, line: pixel coordinates
    or
    x, y: projection coordinates (km or m)
    or
    lon, lat: geographical coordinates in degrees

    Requires
    ========
    gdal to enable sample/line<-> map coords tra'fo's

    >>> p = Point(0,1)

    Need a dataset to try, so I get a MOLA dataset:
    >>> mola = MOLA()
    >>> '%4.2f, %4.2f' % p.pixel_to_meter(mola.dataset.GetGeoTransform())
    '-707109.70, 706994.61'
    >>> p = Point(x=0,y=1)
    >>> '%4.2f, %4.2f' % p.meter_to_pixel(mola.dataset.GetGeoTransform())
    '6144.00, 6143.99'
    >>> p2 = Point(x=3,y=3)
    >>> newP = p + p2
    >>> print(newP.x, newP.y)
    3 4
    """

    @classmethod
    def copy_geodata(cls, point, **kwargs):
        return cls(geotrans=point.geotrans, proj=point.proj, **kwargs)

    def __init__(
        self,
        sample=None,
        line=None,
        x=None,
        y=None,
        lon=None,
        lat=None,
        geotrans=None,
        proj=None,
    ):
        self.sample = sample
        self.line = line
        self.x = x
        self.y = y
        self.lat = lat
        self.lon = lon
        self.centered = False
        self.geotrans = geotrans
        self.proj = proj
        if geotrans is not None:
            if sample is not None:
                self.pixel_to_meter()
            elif x is not None:
                self.meter_to_pixel()
            if proj is not None:
                if sample is not None:
                    self.pixel_to_lonlat()
                elif x is not None:
                    self.lonlat_to_meter()

    @property
    def pixels(self):
        return np.array([self.sample, self.line])

    @property
    def coords(self):
        return np.array([self.x, self.y])

    @property
    def lonlats(self):
        return np.array([self.lon, self.lat])

    def shift_to_center(self, geotransform):
        # if i'd shift, the centerpoint does not show center coordinates
        # so that seems wrong. am i overlooking something?
        # self.x += geotransform[1] / 2.0
        # self.y += geotransform[5] / 2.0
        # self.centered = True
        pass

    def __add__(self, other):
        newPoint = Point(0, 0)
        if all([coord is not None for coord in [self.sample, other.sample]]):
            newPoint.sample = self.sample + other.sample
            newPoint.line = self.line + other.line
        if all([coord is not None for coord in [self.x, other.x]]):
            newPoint.x = self.x + other.x
            newPoint.y = self.y + other.y
        if all([coord is not None for coord in [self.lat, other.lat]]):
            newPoint.lat = self.lat + other.lat
            newPoint.lon = self.lon + other.lon
        return newPoint

    def __str__(self):
        s = f"Pixel: ({self.sample}, {self.line})\n"
        s += f"Map: ({self.x}, {self.y})\n"
        s += f"Lon/Lat: ({self.lon}, {self.lat})"
        return s

    def __repr__(self):
        return self.__str__()

    def pixel_to_meter(self, geotransform=None, shift_to_center=False):
        """provide point in map projection coordinates.

        Parameters
        ==========
        Geotransform in format as given by GDAL datasets.GetGeoTransform()
        If none is provided, the method uses the one given at object creation
        and throws an error if none was given at that time either.
        If a geotransform has been provided, it is copied into the
        object.

        Returns
        =======
        tuple (x,y) coordinates in the projection of the dataset

        >>> p = Point(0,0)
        >>> mola = MOLA()
        >>> '%6.2f, '*2 % p.pixel_to_meter(mola.dataset.GetGeoTransform())
        '-707109.70, 707109.70, '
        """
        # if not given as parameter, take it from self
        if geotransform is None:
            geotransform = self.geotrans
        else:
            # otherwise copy call parameter to self
            self.geotrans = geotransform
        # if it's still None
        if geotransform is None:
            # raise Error, abort method
            raise ProjectionNotSetError("pixel_to_meter")
        if self.sample is None:
            raise SomethingNotSetError("pixel_to_meter", "'sample'")
        self.x, self.y = pixel_to_meter(self.sample, self.line, geotransform)
        return (self.x, self.y)

    def meter_to_pixel(self, geotransform=None):
        """provide pixel coords from x,y coords.

        Input: gdal Dataset
        Return: list [line,sample] of the dataset for given coordinate
        >>> p = Point(x=5e5, y=6e5)
        >>> mola = MOLA()
        >>> '%6.2f ,'*2 % p.meter_to_pixel(mola.dataset.GetGeoTransform())
        '10488.45 ,930.66 ,'
        """
        if geotransform is None:
            geotransform = self.geotrans
        if (self.x is None) or (self.y is None):
            raise SomethingNotSetError(
                (self.x, self.y), "Map coordinates not " "set for transformation."
            )
        tInverse = gdal.InvGeoTransform(geotransform)
        self.sample, self.line = gdal.ApplyGeoTransform(tInverse, self.x, self.y)
        return (self.sample, self.line)

    def pixel_to_lonlat(self, geotransform=None, projection=None):
        self.pixel_to_meter(geotransform)
        self.meter_to_lonlat(projection)
        return (self.lon, self.lat)

    def lonlat_to_pixel(self, geotransform=None, projection=None):
        self.lonlat_to_meter(projection)
        self.meter_to_pixel(geotransform)
        return (self.sample, self.line)

    def meter_to_lonlat(self, projection=None):
        if projection is None:
            projection = self.proj
        if projection is None:
            raise ProjectionNotSetError("lonlat_to_meter")
        srs = debug_srs(projection)
        ct = osr.CoordinateTransformation(srs, srs.CloneGeogCS())
        self.lon, self.lat, height = ct.TransformPoint(self.x, self.y)
        if self.lon < 0:
            self.lon = 360.0 - abs(self.lon)
        return (self.lon, self.lat)

    def lonlat_to_meter(self, projection=None):
        if projection is None:
            projection = self.proj
        if projection is None:
            raise ProjectionNotSetError("lonlat_to_meter")
        srs = debug_srs(projection)
        ct = osr.CoordinateTransformation(srs.CloneGeogCS(), srs)
        # height not used so far!
        self.x, self.y, height = ct.TransformPoint(self.lon, self.lat)
        return (self.x, self.y)

    def calculate_azimuth(self, p2, zero="right"):
        return calculate_image_azimuth(self, p2, zero=zero)

# %% ../notebooks/08_geotools.ipynb 6
class Window:
    """class to manage a window made of corner Points (objects of Point())

    when using width, only quadratic windows supported currently
    >>> p1 = Point(0, 1)
    >>> p2 = Point(10,20)
    """

    def __init__(self, ulPoint=None, lrPoint=None, centerPoint=None, width=None):
        if not any([lrPoint, centerPoint, width]):
            self.usage()
        else:
            self.ul = ulPoint
            self.lr = lrPoint
            self.center = centerPoint
        self.width = width
        if not (ulPoint and lrPoint):
            if centerPoint and width:
                self.get_corners_from_center()
            elif ulPoint and width:
                self.get_lr_from_width()
            else:
                print(
                    "Either upper left and lower right or upper left/"
                    " centerPoint with width needs to be provided."
                )
                return

    def copy(self):
        win = Window(
            Point(
                self.ul.sample,
                self.ul.line,
                x=self.ul.x,
                y=self.ul.y,
                lon=self.ul.lon,
                lat=self.ul.lat,
            ),
            Point(
                self.lr.sample,
                self.lr.line,
                x=self.lr.x,
                y=self.lr.y,
                lon=self.lr.lon,
                lat=self.lr.lat,
            ),
        )
        return win

    def __str__(self):
        return f"{self.ul()}, {self.lr()}, {self.width}, {self.center()}"

    def __repr__(self):
        return self.__str__()

    def get_lr_from_width(self):
        lrSample = self.ul.sample + self.width
        lrLine = self.ul.line + self.width
        self.lr = Point(lrSample, lrLine)
        return self.lr

    def usage(self):
        print(
            """Usage: win = Window(pointObject1, pointObject2)
        or
        win = Window(pointObject1, width_in_Pixel)
        or
        win = Window(centerPoint, width_in_Pixel)"""
        )
        return

    def get_corners_from_center(self):
        """Create symmetric window around given center point.

        Create symmetric window around given sample/line point and return
        to Point objects for each corner point of the window
        >>> win = Window(centerPoint=Point(100,100),width=50)
        >>> win.ul.sample, win.ul.line, win.lr.sample, win.lr.line
        (75, 75, 125, 125)
        """
        ulSample = self.center.sample - self.width // 2
        ulLine = self.center.line - self.width // 2
        self.ul = Point(ulSample, ulLine)
        lrSample = self.center.sample + self.width // 2
        lrLine = self.center.line + self.width // 2
        self.lr = Point(lrSample, lrLine)
        return (self.ul, self.lr)

    def get_gdal_window(self):
        """provide window pixel coordinates in gdal format.

        thats [ulSample, ulLine, samplewidth,linewidth]
        >>> win = Window(Point(10,150),Point(100,200))
        >>> win.get_gdal_window()
        [10, 150, 90, 50]
        """
        return [
            int(self.ul.sample),
            int(self.ul.line),
            int(self.lr.sample - self.ul.sample),
            int(self.lr.line - self.ul.line),
        ]

    def get_extent(self, dataset, lonlat=False):
        """provide window coordinates in matplotlib extent format

        this is needed to get the imshow image plot in the right coordinates
        the strange looking string print out for getting the extent is only
        done to make up for floating point differences that can happen when
        this code executes on other machines
        >>> win = Window(Point(10,150),Point(100,200))
        >>> mola = MOLA()
        >>> '%6.2f, '*4 % tuple(win.get_extent(mola.dataset))
        '-705958.80, -695600.75, 684091.80, 689846.28, '
        """
        if lonlat is False:
            self.ul.pixel_to_meter(dataset.GetGeoTransform())
            self.lr.pixel_to_meter(dataset.GetGeoTransform())
            return [self.ul.x, self.lr.x, self.lr.y, self.ul.y]
        elif lonlat is True:
            self.ul.pixel_to_lonlat(dataset.GetGeoTransform(), dataset.GetProjection())
            self.lr.pixel_to_lonlat(dataset.GetGeoTransform(), dataset.GetProjection())
            return [self.ul.lon, self.lr.lon, self.lr.lat, self.ul.lat]

# %% ../notebooks/08_geotools.ipynb 7
class ImgData:
    """docstring for ImgData"""

    def __init__(self, fname=None):
        self.fname = fname
        self.dataset = gdal.Open(self.fname)
        self.ds = self.dataset
        self.X = self.ds.RasterXSize
        self.Y = self.ds.RasterYSize
        for i in range(self.ds.RasterCount):
            setattr(self, "band" + str(i + 1), self.ds.GetRasterBand(i + 1))
        self.band = self.band1  # keep with older interface of just 1 band
        self.geotrans = self.dataset.GetGeoTransform()
        self.projection = self.dataset.GetProjection()
        self.center = Point(
            self.X // 2, self.Y // 2, geotrans=self.geotrans, proj=self.projection
        )

    def _read_data(self, band):
        band = getattr(self, band)
        data = band.ReadAsArray(*self.window.get_gdal_window())
        ndv = band.GetNoDataValue()
        mdata = np.ma.masked_equal(data, ndv)
        self.data = data
        self.mdata = mdata

    def read_cropped_by_n(self, n):
        "return a window of n x n "
        self.data = self.ds.ReadAsArray(n, n, self.X - n, self.Y - n)

    def read_center_window(self, width=500, band="band1"):
        """Get some sample data from the center of the dataset

        Input: width of square data array, default 500
        >>> mola= MOLA()
        >>> mola.read_center_window()
        >>> mola.data.shape
        (500, 500)
        >>> mola.data.min()
        3382217.5
        >>> mola.data.max()
        3382383.8
        """
        width = min(width, self.X, self.Y)
        self.window = Window(centerPoint=self.center, width=width)
        self._read_data(band)

    def read_all(self, maxdim=1024, band="band1"):
        """This reads all data into a max_dim sized buffer.

        GDAL is doing the downsampling.
        """

        b = getattr(self, band)
        ndv = b.GetNoDataValue()
        ns = self.X
        nl = self.Y

        self.window = Window(ulPoint=Point(0, 0), lrPoint=Point(self.X, self.Y))
        # if full res is smaller than maxdim, get everything
        if (maxdim > self.X) and (maxdim > self.Y):
            data = b.ReadAsArray()
            mdata = np.ma.masked_equal(data, ndv)
        else:
            # Don't want to load the entire dataset for stats computation
            # This is maximum dimension for reduced resolution array
            max_dim = maxdim

            scale_ns = ns / max_dim
            scale_nl = nl / max_dim
            scale_max = max(scale_ns, scale_nl)

            if scale_max > 1:
                nl = round(nl / scale_max)
                ns = round(ns / scale_max)

            # The buf_size parameters determine the final array dimensions
            data = np.array(b.ReadAsArray(buf_xsize=ns, buf_ysize=nl))
            mdata = np.ma.masked_equal(data, ndv)
        self.mdata = mdata
        self.data = data
        return mdata

    def read_window(self, ul_or_win, lrPoint=None, band="band1"):
        """get data for Window object or 2 Point objects

        user can either provide one Window object or 2 Point objects as input
        Point.sample and Point.line have to be filled.
        >>> mola = MOLA()
        >>> data = mola.read_window(Point(100,200),Point(300,400))
        >>> data.shape
        (200, 200)
        >>> data.min()
        3380594.5
        >>> data.max()
        3380967.0
        >>> data = mola.read_window(Window(Point(100,200),Point(400,500)))
        >>> data.shape
        (300, 300)
        >>> data.min()
        3380329.8
        """
        if isinstance(ul_or_win, Window):
            self.window = ul_or_win
        else:
            self.window = Window(ul_or_win, lrPoint)
        self._read_data(band)
        return self.data

    def window_coords_to_meter(self):
        self.window.ul.pixel_to_meter(self.geotrans)
        self.window.lr.pixel_to_meter(self.geotrans)

    def window_coords_to_lonlat(self):
        self.window.ul.pixel_to_lonlat(self.geotrans, self.projection)
        self.window.lr.pixel_to_lonlat(self.geotrans, self.projection)

    def window_coords_to_pixel(self):
        self.window.ul.lonlat_to_pixel(self.geotrans, self.projection)
        self.window.lr.lonlat_to_pixel(self.geotrans, self.projection)

    def normalize(self):
        new_data = self.data - self.data.min()
        new_data /= new_data.max()
        self.data = new_data

    def convert_to_uint8(self):
        self.normalize()
        self.data = np.array(self.data * 256, dtype=np.uint8)
        return self.data

    def show(self, lonlat=False, cb=False):
        fig = figure()
        ax = fig.add_subplot(111)
        extent = self.window.get_extent(self.dataset, lonlat)
        im = ax.imshow(self.data, extent=extent)  # ,origin='image')
        if lonlat:
            ax.set_xlabel("Longitude [deg]")
            ax.set_ylabel("Latitude [deg]")
        else:
            ax.set_xlabel("Rectangular X [m]")
            ax.set_ylabel("Rectangular Y [m]")
        if cb:
            fig.colorbar(im)
        self.ax = ax
        show()

    def add_scalebar(self, loc=3):
        extent = self.window.get_extent(self.dataset)
        diffx = abs(extent[1] - extent[0]) * 1000
        diffy = abs(extent[3] - extent[2]) * 1000
        diff = max(diffx, diffy)
        # get closed magnitude to 10 % of image extent
        scalebarLength = 10 ** int(round(np.log10(diff / 10)))
        scalebarLength /= 1000
        d = dict(
            [
                (1, "1 km"),
                (10, "10 m"),
                (100, "100 m"),
                (1000, "1 km"),
                (10000, "10 km"),
                (100000, "100 km"),
                (1000000, "1000 km"),
            ]
        )
        asb = AnchoredSizeBar(
            self.ax.transData, scalebarLength, d[scalebarLength], loc=loc
        )
        self.ax.add_artist(asb)
        self.ax.get_figure().canvas.draw()



# AUTOGENERATED! DO NOT EDIT! File to edit: ../../notebooks/02c_pds.apps.ipynb.

# %% auto 0
__all__ = ['find_indexes', 'get_index', 'find_instruments', 'refresh_index']

# %% ../../notebooks/02c_pds.apps.ipynb 3
import pandas as pd
from ..config import config
from .indexes import Index

# %% ../../notebooks/02c_pds.apps.ipynb 4
def find_indexes(
    instrument: str,  # Dotted mission.instrument key, e.g. cassini.iss
) -> list:            # List of configured index names
    "Find existing indexes for an instrument."
    return config.list_indexes(instrument)

# %% ../../notebooks/02c_pds.apps.ipynb 6
def get_index(
    instr: str,  # Dotted instrument index, e.g. cassini.iss
    index_name: str='',  # Index name, for exmample 'moon_summary. Optional'
    refresh: bool=False,  # switch to force a refresh of an index
) -> pd.DataFrame:  # The PDS index convert to pandas DataFrame
    """Example: get_index("cassini.iss", "index")"""
    if not index_name:
        index = Index(instr)
    else:
        index = Index(instr + ".indexes." + index_name)
    if index.needs_download or refresh is True:
        print("Index is outdated or not existing or asked to be refreshed.\nDownloading...")
        index.download(force_update=refresh)
        index.convert_to_parquet()
    try:
        return index.parquet
    except FileNotFoundError:
        index.convert_to_parquet()
        return index.parquet

# %% ../../notebooks/02c_pds.apps.ipynb 12
def find_instruments(
    mission: str,  # Mission string, e.g. "cassini"
) -> list:  # List of configured instrument names
    "Find existing instruments for a mission."
    return config.list_instruments(mission)

# %% ../../notebooks/02c_pds.apps.ipynb 15
def refresh_index(
    instr: str,  # Dotted instrument index, e.g. cassini.iss
    index_name: str,  # Index name, for exmample 'moon_summary'
    download: bool = False,  # also refresh dowload, not only local parquet conversion
):
    """Example: refresh_index("cassini.iss", "index")"""
    index = Index(instr + ".indexes." + index_name)
    if download:
        index.download(force_update=True)
    else:
        index.convert_to_parquet()
    return index.parquet



# AUTOGENERATED! DO NOT EDIT! File to edit: ../../notebooks/02a_pds.indexes.ipynb.

# %% auto 0
__all__ = ['logger', 'storage_root', 'dynamic_urls', 'Index']

# %% ../../notebooks/02a_pds.indexes.ipynb 3
import logging
from datetime import datetime
from pathlib import Path
from urllib.parse import urlsplit, urlunsplit
from urllib.request import URLError

import tomlkit as toml
from dateutil import parser

import pandas as pd
from .. import utils
from ..config import config
from .ctx_index import CTXIndex
from .lroc_index import LROCIndex
from .utils import IndexLabel, fix_hirise_edrcumindex

logger = logging.getLogger(__name__)

storage_root = Path(config.storage_root)

# %% ../../notebooks/02a_pds.indexes.ipynb 4
dynamic_urls = {"mro.ctx": CTXIndex, "lro.lroc": LROCIndex}

# %% ../../notebooks/02a_pds.indexes.ipynb 6
class Index:
    """Index manager class.

    This class manages one index, identified by a dotted key, e.g. `cassini.iss.ring_summary`
    """

    def __init__(
        self,
        key: str,  # Nested (dotted) key, e.g. cassini.iss.ring_summary
        url: str = None,  # URL to index. If not given, will be read from config object.
    ):
        self.key = self.parse_key(key)
        self.set_url(url)
        try:
            self.timestamp = parser.parse(config.get_value(self.key)["timestamp"])
        except toml.exceptions.NonExistentKey:
            if self.local_label_path.exists():
                self.timestamp = datetime.fromtimestamp(
                    self.local_label_path.stat().st_mtime
                )
                self.update_timestamp()
            else:
                self.timestamp = None
        self.new_timestamp = None  # filled by needs_download()

    def parse_key(
        self,
        key: str,  # dotted key
    ):
        """Take care of different ways how the key could be structured.

        This involves adding the sub-key `indexes` for the config file structure,
        which is something the user of this class should not need to know.
        """
        tmp = key if key.startswith("missions") else "missions." + key
        subs = tmp.split(".")
        if subs[3] != "indexes":
            subs.insert(3, "indexes")
        return ".".join(subs)

    def set_url(self, url):  # URL to index.
        "Set URL from having it dynamically determined (for non-static index URLs)."
        self.url = config.get_value(self.key)["url"] if url is None else url
        if not self.url:  # empty ''
            self.url = dynamic_urls[self.instrument_key]().latest_index_label_url

    @property
    def isotimestamp(self):
        return self.timestamp.isoformat()

    @property
    def needs_download(self)->bool:  # Boolean indicating if download is required
        """Property indicating if the index needs to be downloaded.

        Need is True when
        (1) no local timestamp was stored or
        (2) when the remote timestamp is newer.
        """
        try:
            self.remote_timestamp = utils.get_remote_timestamp(self.url)
        except URLError:
            return None
        if self.timestamp:
            if self.remote_timestamp > self.timestamp:
                return True
            else:
                return False
        else:
            # also return True when the timestamp is not valid
            return True

    @property
    def key_tokens(self):
        return self.key.split(".")

    @property
    def mission(self):
        return self.key_tokens[1]

    @property
    def mission_key(self):
        return ".".join(self.key_tokens[1:2])

    @property
    def instrument(self):
        return self.key_tokens[2]

    @property
    def instrument_key(self):
        return ".".join(self.key_tokens[1:3])

    @property
    def index_name(self):
        "str: Examples: EDR, RDR, moon_summary"
        return self.key_tokens[3]

    @property
    def label_filename(self):
        return Path(self.url.split("/")[-1])

    @property
    def isupper(self):
        return self.label_filename.suffix.isupper()

    @property
    def table_filename(self):
        new_suffix = ".TAB" if self.isupper else ".tab"
        return self.label_filename.with_suffix(new_suffix)

    @property
    def label_path(self):
        return Path(urlsplit(self.url).path)

    @property
    def table_path(self):
        return self.label_path.with_name(self.table_filename.name)

    @property
    def table_url(self):
        tokens = urlsplit(self.url)
        return urlunsplit(
            tokens._replace(
                path=str(self.label_path.with_name(self.table_filename.name))
            )
        )

    @property
    def local_dir(self):
        p = storage_root / str(self.key).replace(".", "/")
        p.mkdir(parents=True, exist_ok=True)
        return p

    @property
    def local_table_path(self):
        return self.local_dir / self.table_filename

    @property
    def local_label_path(self):
        return self.local_dir / self.label_filename

    @property
    def local_hdf_path(self):
        return self.local_table_path.with_suffix(".hdf")

    @property
    def local_parq_path(self):
        return self.local_table_path.with_suffix(".parq")

    @property
    def df(self):
        return pd.read_hdf(self.local_hdf_path)

    @property
    def parquet(self):
        return pd.read_parquet(self.local_parq_path)

    def download(
        self,
        convert_to_hdf: bool = False,  # Switch to enable conversion to HDF
        convert_to_parquet: bool = True,  # Switch to enable conversion to parquet
        force_update: bool = False,  # Switch to enable a fresh download and conversion
    ):
        """Wrapping URLs for downloading PDS indices and their label files."""
        # check timestamp
        ret = self.needs_download
        if ret is None:
            print("Could not check for any index updates, maybe server is offline?")
            return
        if not ret and not force_update:
            print("Stored index is up-to-date.")
            return
        label_url = self.url
        logger.info("Downloading %s." % label_url)
        utils.url_retrieve(label_url, self.local_label_path)
        logger.info("Downloading %s.", self.table_url)
        utils.url_retrieve(self.table_url, self.local_table_path)
        print(f"Downloaded {self.local_label_path} and {self.local_table_path}")
        if self.key == 'missions.mro.hirise.indexes.edr':  # HiRISE EDR index is broken on the PDS. Team knows.
            print("Fixing broken EDR index...")
            fix_hirise_edrcumindex(self.local_table_path, self.local_table_path.with_name("temp.tab"))
            self.local_table_path.with_name("temp.tab").rename(self.local_table_path)
        self.timestamp = self.remote_timestamp
        self.update_timestamp()
        
        if convert_to_hdf:
            try:
                self.convert_to_hdf()
            except:  # any conversion error simpy leads to HDF marked as missing
                self.set_hdf_available(False)
            else:
                self.set_hdf_available(True)
                print(f"Converted to pandas HDF:\n{self.local_hdf_path}")
        elif convert_to_parquet:
            try:
                self.convert_to_parqet()
            except: 
                self.set_parquet_available(False)
            else:
                self.set_parquet_available(True)
                print(f"Converted to parquet:\n{self.local_parq_path}")


    def set_hdf_available(self, status):
        config.set_value(f"{self.key}.hdf_available", status)

    def set_parquet_available(self, status):
        config.set_value(f"{self.key}.parquet_available", status)

    def update_timestamp(self):
        # Note: the config object writes itself out after setting any value
        config.set_value(f"{self.key}.timestamp", self.isotimestamp)

    @property
    def label(self):
        return IndexLabel(self.local_label_path)

    def read_index_data(self):
        df = self.label.read_index_data()
        return df

    def convert_to_hdf(self):
        df = self.read_index_data()
        df.to_hdf(self.local_hdf_path, "df")

    def convert_to_parquet(self):
        df = self.read_index_data()
        df = df.convert_dtypes()
        df.to_parquet(self.local_parq_path)

    def __str__(self):
        s = f"Key: {self.key}\n"
        s += f"URL: {self.url}\n"
        s += f"Timestamp: {self.timestamp}\n"
        return s

    def __repr__(self):
        return self.__str__()



# AUTOGENERATED! DO NOT EDIT! File to edit: ../notebooks/00_config.ipynb.

# %% auto 0
__all__ = ['config', 'Config']

# %% ../notebooks/00_config.ipynb 3
import os
import shutil
from functools import reduce
from importlib.resources import path as resource_path
from typing import Union

import tomlkit as toml
from fastcore.utils import AttrDict, Path

# %% ../notebooks/00_config.ipynb 5
class Config:
    """Manage config stuff.

    Attributes
    -------
    path: pathlib.Path

    The key, value pairs found in the config file become attributes of the
    class instance after initialization.
    At minimum, there should be the `storage_root` attribute for storing data
    for this package.
    """

    # This part enables a config path location override using env PYCISS_CONFIG
    fname = "planetarypy_config.toml"
    # separating fname from fpath so that resource_path below is correct.
    path = Path(os.getenv("PLANETARYPY_CONFIG", Path.home() / f".{fname}"))

    def __init__(
        self, 
        config_path:str=None  # str or pathlib.Path
    ):
        "Switch to other config file location with `config_path`."
        if config_path is not None:
            self.path = Path(config_path)
        if not self.path.exists():
            with resource_path("planetarypy.data", self.fname) as p:
                shutil.copy(p, self.path)
        self.read_config()
        self.update_missions()

    def read_config(self):
        """Read the configfile and store config dict.

        `storage_root` will be stored as attribute.
        """
        self.tomldoc = toml.loads(self.path.read_text())
        if not self.tomldoc["storage_root"]:
            path = Path.home() / "planetarypy_data"
            path.mkdir()
            self.tomldoc["storage_root"] = str(path)
            self.storage_root = path
            self.save()
        else:
            self.storage_root = Path(self.tomldoc["storage_root"])

    @property
    def d(self):
        "get the Python dic from"
        return self.tomldoc

    def get_value(
        self,
        key:str  # A nested key in dotted format, e.g. cassini.uvis.indexes
    ):
        """Get sub-dictionary by nested key."""
        if not key.startswith('missions'):
            key = 'missions.' + key
        try:
            return reduce(lambda c, k: c[k], key.split("."), self.d)
        except toml.exceptions.NonExistentKey:
            return None

    def set_value(
        self,
        nested_key:str,  # A nested key in dotted format, e.g. cassini.uvis.ring_summary 
        value:Union[float, str],  # Value for the given key to be stored
        save:bool=True  # Switch to control writing out to disk
    ):
        "Set value in sub-dic using dotted key."
        dic = self.tomldoc
        keys = nested_key.split(".")
        for key in keys[:-1]:
            dic = dic[key]
        dic[keys[-1]] = value
        if save:
            self.save()

    def save(self):
        "Write the TOML doc to file."
        self.path.write_text(toml.dumps(self.tomldoc))

    def ask_storage_root(self):
        """Use input() to ask user for the storage_root path.

        The path will be stored in the TOML-dict and saved into existing config file
        at `Class.path`, either default or as given during init.
        `storage_root` attribute is set as well.
        """
        path = input(
            "Provide the root storage path where all downloaded and produced data will be stored:"
        )
        self.tomldoc["storage_root"] = path
        self.storage_root = Path(path)
        self.save()

    @property
    def missions(self):
        return list(self.d["missions"].keys())

    def list_instruments(self, mission):
        if not mission.startswith("missions"):
            mission = "missions." + mission
        instruments = self.get_value(mission)
        return list(instruments.keys())

    def list_indexes(self, instrument):
        "instrument key needs to be <mission>.<instrument>"
        if not instrument.startswith("missions"):
            instrument = "missions." + instrument
        indexes = self.get_value(instrument + ".indexes")
        return list(indexes)

    def _copy_clean_to_resource(self):
        "Copy a clean config file without timestamps into resource path for repo commit."
        dic = self.tomldoc.copy()
        missions = dic["missions"]
        for mission in missions.keys():
            mdict = missions[mission]
            for instr in mdict.keys():
                instrdict = mdict[instr]
                for index in instrdict["indexes"]:
                    instrdict["indexes"][index]["timestamp"] = ""
        with resource_path("planetarypy.data", self.fname) as p:
            Path(p).write_text(toml.dumps(dic))

    def update_missions(self):
        "Check if a new version with more URLs exist at resource path."
        with resource_path("planetarypy.data", self.fname) as p:
            new = toml.loads(Path(p).read_text())["missions"]
        old = self.tomldoc["missions"]
        for mission in new:
            missiondata = new[mission]
            if mission not in old:
                old[mission] = missiondata
                continue
            for instr in missiondata:
                instrdata = missiondata[instr]
                if instr not in old[mission]:
                    old[mission][instr] = instrdata
                    continue
                for index in instrdata["indexes"]:
                    indexdata = instrdata["indexes"][index]
                    if index not in old[mission][instr]["indexes"]:
                        old[mission][instr]["indexes"][index] = indexdata
                        continue
                    oldindexdata = old[mission][instr]["indexes"][index]
                    if indexdata["url"] != oldindexdata["url"]:
                        oldindexdata["url"] = indexdata["url"]
        self.save()

    def populate_timestamps(self):
        pass

    def __repr__(self):
        return AttrDict(self.d).__repr__()

# %% ../notebooks/00_config.ipynb 6
config = Config()

# %% ../notebooks/00_config.ipynb 10
config = Config()



# AUTOGENERATED! DO NOT EDIT! File to edit: ../../notebooks/02e_pds.lroc_index.ipynb.

# %% auto 0
__all__ = ['LROCIndex']

# %% ../../notebooks/02e_pds.lroc_index.ipynb 2
import warnings
from nbdev.showdoc import *

# %% ../../notebooks/02e_pds.lroc_index.ipynb 3
class LROCIndex:
    def __init__(self):
        warnings.warn("LROCIndex not implemented yet.")



# AUTOGENERATED! DO NOT EDIT! File to edit: ../notebooks/07_diviner.ipynb.

# %% auto 0
__all__ = ['hostname', 'headerstring', 'DataManager', 'parse_header_line', 'L1AHeader', 'read_l1a_data', 'get_data_path']

# %% ../notebooks/07_diviner.ipynb 3
import socket
from pathlib import Path

import hvplot.xarray  # noqa
import pandas as pd
from .config import config
from .utils import url_retrieve
from yarl import URL

# %% ../notebooks/07_diviner.ipynb 4
hostname = socket.gethostname()
if hostname.startswith("luna") and hostname.endswith("diviner.ucla.edu"):
    storage_root = Path("/q/marks/feidata/DIV:opsL1A/data")
else:
    storage_root = config.storage_root / "missions/lro/diviner"

# %% ../notebooks/07_diviner.ipynb 5
class DataManager:
    def __init__(self, tstr):
        self.tstr = tstr

    @property
    def year(self):
        return self.tstr[:4]

    @property
    def yearmonth(self):
        return self.tstr[:6]

    @property
    def yearmonthday(self):
        return self.tstr[:8]

    @property
    def volume(self):
        if int(self.year) <= 2016:
            vol = 1
        else:
            vol = 2
        return vol

    @property
    def base_url(self):
        return URL(
            f"https://pds-geosciences.wustl.edu/lro/lro-l-dlre-2-edr-v1/lrodlr_{str(self.volume).zfill(4)}/data/"
        )

    @property
    def url(self):
        return self.base_url / str(
            Path(self.year, self.yearmonth, self.yearmonthday, f"{self.tstr}_edr.tab")
        )

    @property
    def local_path(self):
        end = self.url.parts[-4:]
        return storage_root / Path(*end)

    def download(self, overwrite=False):
        if self.local_path.exists() and not overwrite:
            print("File exists. Use `overwrite=True` to get a fresh copy.")
            return
        self.local_path.parent.mkdir(parents=True, exist_ok=True)
        url_retrieve(self.url, self.local_path)

# %% ../notebooks/07_diviner.ipynb 8
headerstring = (
    "Q, DATE, UTC, SCLK, SOUNDING, FROM_PKT, PKT_COUNT, SAFING, SAFED, FREEZING, FROZEN, ROLLING, DUMPING, MOVING, TEMP_FAULT,   SC_TIME_SECS,   SC_TIME_SUBS, TICKS_PKT_START, TICKS_AT_SC_TIME, OST_INDEX, EST_INDEX, SST_INDEX, LAST_AZ_CMD, LAST_EL_CMD, FPA_TEMP, FPB_TEMP, BAFFLE_A_TEMP, BAFFLE_B_TEMP, BB_1_TEMP, OBA_1_TEMP, ERROR_TIME, ERROR_ID,  ERROR_DETAIL , ERROR_COUNT, COMMANDS_RECEIVED, COMMANDS_EXECUTED, COMMANDS_REJECTED,    LAST_COMMAND_REC ,      CMD,  REQ_ID , LAST_TIME_COMMAND, LAST_EQX_PREDICTION, HYBRID_TEMP, FPA_TEMP_CYC, FPB_TEMP_CYC, BAFFLE_A_TEMP_CYC, BAFFLE_B_TEMP_CYC, OBA_1_TEMP_CYC, OBA_2_TEMP, BB_1_TEMP_CYC, BB_2_TEMP, SOLAR_TARGET_TEMP, YOKE_TEMP, EL_ACTUATOR_TEMP, AZ_ACTUATOR_TEMP,  MIN_15V, PLU_15V, SOLAR_BASE_TEMP, PLU_5V, "
    "A1_01, A1_02, A1_03, A1_04, A1_05, A1_06, A1_07, A1_08, A1_09, A1_10, A1_11, A1_12, A1_13, A1_14, A1_15, A1_16, A1_17, A1_18, A1_19, A1_20, A1_21, A2_01, A2_02, A2_03, A2_04, A2_05, A2_06, A2_07, A2_08, A2_09, A2_10, A2_11, A2_12, A2_13, A2_14, A2_15, A2_16, A2_17, A2_18, A2_19, A2_20, A2_21, A3_01, A3_02, A3_03, A3_04, A3_05, A3_06, A3_07, A3_08, A3_09, A3_10, A3_11, A3_12, A3_13, A3_14, A3_15, A3_16, A3_17, A3_18, A3_19, A3_20, A3_21, A4_01, A4_02, A4_03, A4_04, A4_05, A4_06, A4_07, A4_08, A4_09, A4_10, A4_11, A4_12, A4_13, A4_14, A4_15, A4_16, A4_17, A4_18, A4_19, A4_20, A4_21, A5_01, A5_02, A5_03, A5_04, A5_05, A5_06, A5_07, A5_08, A5_09, A5_10, A5_11, A5_12, A5_13, A5_14, A5_15, A5_16, A5_17, A5_18, A5_19, A5_20, A5_21, A6_01, A6_02, A6_03, A6_04, A6_05, A6_06, A6_07, A6_08, A6_09, A6_10, A6_11, A6_12, A6_13, A6_14, A6_15, A6_16, A6_17, A6_18, A6_19, A6_20, A6_21, B1_01, B1_02, B1_03, B1_04, B1_05, B1_06, B1_07, B1_08, B1_09, B1_10, B1_11, B1_12, B1_13, B1_14, B1_15, B1_16, B1_17, B1_18, B1_19, B1_20, B1_21, B2_01, B2_02, B2_03, B2_04, B2_05, B2_06, B2_07, B2_08, B2_09, B2_10, B2_11, B2_12, B2_13, B2_14, B2_15, B2_16, B2_17, B2_18, B2_19, B2_20, B2_21, B3_01, B3_02, B3_03, B3_04, B3_05, B3_06, B3_07, B3_08, B3_09, B3_10, B3_11, B3_12, B3_13, B3_14, B3_15, B3_16, B3_17, B3_18, B3_19, B3_20, B3_21"
)


def parse_header_line(line):
    """Parse header lines.

    >>> s = ' a   b  c    '
    >>> parse_header_line(s)
    ['a', 'b', 'c']
    >>> s = '  a, b  ,   c '
    >>> parse_header_line(s)
    ['a', 'b', 'c']
    """
    line = line.strip("#")
    if "," in line:
        newline = line.split(",")
    else:
        newline = line.split()
    return [i.strip().lower() for i in newline]


class L1AHeader:
    # beware: parse_header_line converts to lower case!
    columns = parse_header_line(headerstring)

    tel1cols = [
        "a{0}_{1}".format(i, str(j).zfill(2)) for i in range(1, 7) for j in range(1, 22)
    ]
    tel2cols = [
        "b{0}_{1}".format(i, str(j).zfill(2)) for i in range(1, 4) for j in range(1, 22)
    ]

    datacols = tel1cols + tel2cols

    metadatacols = list(set(columns) - set(datacols))
    metadatacols.sort()


def read_l1a_data(fname, nrows=None):
    df = pd.io.parsers.read_csv(
        fname,
        names=L1AHeader.columns,
        na_values="-9999",
        skiprows=8,
        skipinitialspace=True,
    )
    return df
    return parse_times(df)

# %% ../notebooks/07_diviner.ipynb 17
def get_data_path(tstr):
    dm = DataManager(tstr)
    if not dm.local_path.exists():
        dm.download()
    return dm.local_path



# AUTOGENERATED! DO NOT EDIT! File to edit: ../notebooks/09_exceptions.ipynb.

# %% auto 0
__all__ = ['Error', 'SomethingNotSetError', 'ProjectionNotSetError', 'GeoTransformNotSetError', 'KMASpiceError',
           'SPointNotSetError', 'ObserverNotSetError']

# %% ../notebooks/09_exceptions.ipynb 2
class Error(Exception):
    """Base class for exceptions in this module."""

    pass


class SomethingNotSetError(Error):
    """Exception raised for errors in the input of transformations.

    Attributes:
        where -- where is something missing
        what     -- what is missing
    """

    def __init__(self, where, what):
        self.where = where
        self.what = what

    def __str__(self):
        return "{0} not set in {1}".format(self.what, self.where)


class ProjectionNotSetError(SomethingNotSetError):
    what = "Projection"


class GeoTransformNotSetError(SomethingNotSetError):
    what = "GeoTransform"


class KMASpiceError(Exception):
    """Base class for exceptions in this module."""

    pass


class SPointNotSetError(KMASpiceError):
    def __str__(self):
        return """You are trying to use a method that requires that the surface
point is defined. The class member is <spoint>. It can be set using the method
'set_spoint_by'. This operation had no effect."""


class ObserverNotSetError(KMASpiceError):
    def __str__(self):
        return """The method you called requires an observer to be set.
                  This operation had no effect."""



# AUTOGENERATED! DO NOT EDIT! File to edit: ../../notebooks/02d_pds.opusapi.ipynb.

# %% auto 0
__all__ = ['base_url', 'metadata_url', 'image_url', 'dic', 'OPUSImageURL', 'OPUSObsID', 'OPUS']

# %% ../../notebooks/02d_pds.opusapi.ipynb 2
from pathlib import Path
from urllib.parse import urlencode, urlparse
from urllib.request import unquote, urlretrieve

import pandas as pd
import requests
from IPython.display import HTML, display

base_url = "https://opus.pds-rings.seti.org/opus/api"
metadata_url = base_url + "/metadata"
image_url = base_url + "/image/"

dic = {"raw_data": "coiss_raw", "calibrated_data": "coiss_calib"}




class OPUSImageURL(object):

    """Manage URLS from the OPUS response."""

    def __init__(self, jsonlist):
        self.jsonlist = jsonlist
        for item in jsonlist:
            parsed = urlparse(item)
            if "//" in parsed.path:
                continue
            if item.upper().endswith(".LBL"):
                self.label_url = item
            elif item.upper().endswith(".IMG"):
                self.image_url = item

    def __repr__(self):
        s = "Label:\n{}\nImage:\n{}".format(self.label_url, self.image_url)
        return s


class OPUSObsID(object):

    """Manage observation IDs from OPUS responses."""

    def __init__(self, obsid_data):
        self.idname = obsid_data[0]
        self.raw = OPUSImageURL(obsid_data[1][dic["raw_data"]])
        # the images have an iteration number. I'm fishing it out here:
        self.number = self.raw.image_url.split("_")[-1][0]
        try:
            self.calib = OPUSImageURL(obsid_data[1][dic["calibrated_data"]])
        except KeyError:
            self.calib = None

    def _get_img_url(self, size):
        base = self.raw.label_url[:-4].replace("volumes", "browse")
        return "{}_{}.jpg".format(base, size)

    @property
    def raw_urls(self):
        return [self.raw.image_url, self.raw.label_url]

    @property
    def calib_urls(self):
        return [self.calib.image_url, self.calib.label_url]

    @property
    def all_urls(self):
        return self.raw_urls + self.calib_urls

    @property
    def img_id(self):
        """Convert OPUS ObsID into the more known image_id."""
        tokens = self.idname.split("-")
        return tokens[-1]

    @property
    def small_img_url(self):
        return self._get_img_url("small")

    @property
    def medium_img_url(self):
        return self._get_img_url("med")

    @property
    def thumb_img_url(self):
        return self._get_img_url("thumb")

    @property
    def full_img_url(self):
        return self._get_img_url("full")

    def get_meta_data(self):
        return MetaData(self.img_id)

    def __repr__(self):
        s = "Raw:\n{}\nCalibrated:\n{}".format(self.raw, self.calib)
        return s


class OPUS(object):

    """Manage OPUS API requests.


    """

    def __init__(self, silent=False):
        self.silent = silent

    def query_image_id(self, image_id):
        """Query OPUS via the image_id.

        This is a query using the 'primaryfilespec' field of the OPUS database.
        It returns a list of URLS into the `obsids` attribute.

        This example queries for an image of Titan:

        >>> opus = opusapi.OPUS()
        >>> opus.query_image_id('N1695760475_1')

        After this, one can call `download_results()` to retrieve the found
        data into the standard locations into the database_path as defined in
        `.pyciss.yaml` (the config file),
        """
        myquery = {"primaryfilespec": image_id}
        self.create_files_request(myquery, fmt="json")
        self.unpack_json_response()
        return self.obsids

    def create_request_with_query(self, kind, query, size="thumb", fmt="json"):
        """api/data.[fmt], api/images/[size].[fmt] api/files.[fmt]

        kind = ['data', 'images', 'files']


        """
        if kind == "data" or kind == "files":
            url = "{}/{}.{}".format(base_url, kind, fmt)
        elif kind == "images":
            url = "{}/images/{}.{}".format(base_url, size, fmt)
        self.url = url
        self.r = requests.get(url, params=unquote(urlencode(query)))

    def create_files_request(self, query, fmt="json"):
        self.create_request_with_query("files", query, fmt=fmt)

    def create_images_request(self, query, size="thumb", fmt="html"):
        self.create_request_with_query("images", query, size=size, fmt=fmt)

    def get_volume_id(self, ring_obsid):
        url = "{}/{}.json".format(metadata_url, ring_obsid)
        query = {"cols": "volumeidlist"}
        r = requests.get(url, params=unquote(urlencode(query)))
        return r.json()[0]["volume_id_list"]

    # def create_data_request(self, query, fmt='json'):
    #     myquery = query.copy()
    #     myquery.update(query)
    #     self.create_request_with_query('data', myquery, fmt=fmt)

    @property
    def response(self):
        return self.r.json()["data"]

    def unpack_json_response(self):
        if self.r.status_code == 500:
            if not self.silent:
                print("No data found.")
            self.obsids = []
            return
        obsids = []
        for obsid_data in self.response.items():
            obsids.append(obsid_data)
        self.obsids = obsids
        if not self.silent:
            print("Found {} obsids.".format(len(obsids)))
            if len(obsids) == 1000:
                print(
                    "List is 1000 entries long, which is the pre-set limit, hence"
                    " the real number of results might be longe."
                )

    def get_radial_res_query(self, res1, res2):
        myquery = dict(
            target="S+RINGS",
            instrumentid="Cassini+ISS",
            projectedradialresolution1=res1,
            projectedradialresolution2=res2,
            limit=1000,
        )
        return myquery

    def _get_time_query(self, t1, t2):
        myquery = dict(instrumentid="Cassini+ISS", timesec1=t1, timesec2=t2)
        return myquery

    def get_between_times(self, t1, t2, target=None):
        """
        Query for OPUS data between times t1 and t2.

        Parameters
        ----------
        t1, t2 : datetime.datetime, strings
            Start and end time for the query. If type is datetime, will be
            converted to isoformat string. If type is string already, it needs
            to be in an accepted international format for time strings.
        target : str
            Potential target for the observation query. Most likely will reduce
            the amount of data matching the query a lot.

        Returns
        -------
        None, but set's state of the object to have new query results stored
        in self.obsids.
        """
        try:
            # checking if times have isoformat() method (datetimes have)
            t1 = t1.isoformat()
            t2 = t2.isoformat()
        except AttributeError:
            # if not, should already be a string, so do nothing.
            pass
        myquery = self._get_time_query(t1, t2)
        if target is not None:
            myquery["target"] = target
        self.create_files_request(myquery, fmt="json")
        self.unpack_json_response()

    def get_between_resolutions(self, res1="", res2="0.5"):
        myquery = self.get_radial_res_query(res1, res2)
        self.create_files_request(myquery, fmt="json")
        self.unpack_json_response()

    def show_images(self, size="small"):
        """Shows preview images using the Jupyter notebook HTML display.

        Parameters
        ==========
        size : {'small', 'med', 'thumb', 'full'}
            Determines the size of the preview image to be shown.
        """
        d = dict(small=256, med=512, thumb=100, full=1024)
        try:
            width = d[size]
        except KeyError:
            print("Allowed keys:", d.keys())
            return
        img_urls = [i._get_img_url(size) for i in self.obsids]
        imagesList = "".join(
            [
                "<img style='width: {0}px; margin: 0px; float: "
                "left; border: 1px solid black;' "
                "src='{1}' />".format(width, s)
                for s in img_urls
            ]
        )
        display(HTML(imagesList))

    def download_results(self, savedir=None, raw=True, calib=False, index=None):
        """Download the previously found and stored Opus obsids.

        Parameters
        ==========
        savedir: str or pathlib.Path, optional
            If the database root folder as defined by the config.ini should not be used,
            provide a different savedir here. It will be handed to PathManager.
        """
        obsids = self.obsids if index is None else [self.obsids[index]]
        for obsid in obsids:
            pm = io.PathManager(obsid.img_id, savedir=savedir)
            pm.basepath.mkdir(exist_ok=True)
            to_download = []
            if raw is True:
                to_download.extend(obsid.raw_urls)
            if calib is True:
                to_download.extend(obsid.calib_urls)
            for url in to_download:
                basename = Path(url).name
                print("Downloading", basename)
                store_path = str(pm.basepath / basename)
                try:
                    urlretrieve(url, store_path)
                except Exception as e:
                    urlretrieve(url.replace("https", "http"), store_path)
            return str(pm.basepath)

    def download_previews(self, savedir=None):
        """Download preview files for the previously found and stored Opus obsids.

        Parameters
        ==========
        savedir: str or pathlib.Path, optional
            If the database root folder as defined by the config.ini should not be used,
            provide a different savedir here. It will be handed to PathManager.
        """
        for obsid in self.obsids:
            pm = io.PathManager(obsid.img_id, savedir=savedir)
            pm.basepath.mkdir(exist_ok=True)
            basename = Path(obsid.medium_img_url).name
            print("Downloading", basename)
            urlretrieve(obsid.medium_img_url, str(pm.basepath / basename))




# AUTOGENERATED! DO NOT EDIT! File to edit: ../notebooks/06_cassini_iss.ipynb.

# %% auto 0
__all__ = ['storage_root', 'opus_keys', 'ISSOpus']

# %% ../notebooks/06_cassini_iss.ipynb 2
from yarl import URL

from .config import config
from .pds.opusapi import OPUS
from .utils import have_internet, url_retrieve

# %% ../notebooks/06_cassini_iss.ipynb 3
storage_root = config.storage_root / "missions/cassini/iss"
opus_keys = [
    "coiss_raw",
    "coiss_calib",
    "coiss_thumb",
    "coiss_medium",
    "coiss_full",
    "rms_index",
    "inventory",
    "planet_geometry",
    "ring_geometry",
    "browse_thumb",
    "browse_small",
    "browse_medium",
    "browse_full",
]

# %% ../notebooks/06_cassini_iss.ipynb 5
class ISSOpus:
    def __init__(self, pid):
        self.pid = pid
        if have_internet():
            self.query_pid(pid)

    def query_pid(self, pid):
        opus = OPUS()
        self.query_result = opus.query_image_id(pid)[0]
        self.id = self.query_result[0]
        self.dict = self.query_result[1]
        for k, v in self.dict.items():
            if isinstance(v, list) and len(v) == 1:
                setattr(self, k, v[0])
            else:
                setattr(self, k, v)

    @property
    def raw_data_url(self):
        return URL(self.coiss_raw[0])

    @property
    def raw_label_url(self):
        return URL(self.coiss_raw[1])

    @property
    def raw_prefix_fmt_url(self):
        return URL(self.coiss_raw[2])

    @property
    def raw_tlmtab_url(self):
        return URL(self.coiss_raw[3])

    @property
    def calib_data_url(self):
        return URL(self.coiss_calib[0])

    @property
    def calib_label_url(self):
        return URL(self.coiss_calib[1])

    @property
    def volume(self):
        return self.raw_data_url.parts[4]

    @property
    def local_folder(self):
        return storage_root / self.pid.upper()

    @property
    def local_data_path(self):
        return self.local_folder / self.raw_data_url.name

    @property
    def local_calib_path(self):
        return self.local_folder / self.calib_data_url.name

    @property
    def local_label_path(self):
        return self.local_data_path.with_suffix(".LBL")

    @property
    def local_calib_label_path(self):
        return self.local_calib_path.with_suffix(".LBL")

    def download_raw(self, overwrite=False):
        self.local_data_path.parent.mkdir(parents=True, exist_ok=True)
        if self.local_data_path.exists() and not overwrite:
            print("File exists. Use `overwrite=True` to download fresh.")
            return
        url_retrieve(self.raw_data_url, self.local_data_path)
        url_retrieve(self.raw_label_url, self.local_label_path)

    def download_calib(self, overwrite=False):
        self.local_calib_path.parent.mkdir(parents=True, exist_ok=True)
        if self.local_calib_path.exists():
            if not overwrite:
                print("File exists. Use `overwrite` to force re-download.")
                return
        url_retrieve(self.calib_data_url, self.local_calib_path)

    def __repr__(self):
        s = f"Product ID:\n{self.id}\n\n"
        for k, v in self.query_result[1].items():
            s += f"Key: {k},\nValue(s):\n{v}\n\n"
        return s



# AUTOGENERATED! DO NOT EDIT! File to edit: ../notebooks/03_ctx.ipynb.

# %% auto 0
__all__ = ['baseurl', 'storage_root', 'edrindex', 'catch_isis_error', 'CTXEDR', 'CTX', 'ctx_calib', 'CTXEDRCollection']

# %% ../notebooks/03_ctx.ipynb 3
import warnings
from pathlib import Path

import hvplot.xarray  # noqa
import rasterio
import rioxarray as rxr
from dask import compute, delayed
from .config import config
from .pds.apps import get_index
from .utils import file_variations, url_retrieve
from tqdm.auto import tqdm
from yarl import URL
from fastcore.script import call_parse
from fastcore.basics import store_attr

try:
    from kalasiris.pysis import (
        ProcessError,
        ctxcal,
        ctxevenodd,
        getkey,
        mroctx2isis,
        spiceinit,
    )
except KeyError:
    warnings.warn("kalasiris has a problem initializing ISIS")

# %% ../notebooks/03_ctx.ipynb 4
warnings.filterwarnings("ignore", category=rasterio.errors.NotGeoreferencedWarning)
baseurl = URL(
    "https://pds-imaging.jpl.nasa.gov/data/mro/mars_reconnaissance_orbiter/ctx/"
)

storage_root = config.storage_root / "missions/mro/ctx"
edrindex = get_index("mro.ctx", "edr")

# %% ../notebooks/03_ctx.ipynb 6
def catch_isis_error(func):
    def inner(*args, **kwargs):
        try:
            func(*args, **kwargs)
        except ProcessError as err:
            print("Had ISIS error:")
            print(" ".join(err.cmd))
            print(err.stdout)
            print(err.stderr)

    return inner

# %% ../notebooks/03_ctx.ipynb 7
class CTXEDR:
    """Manage access to EDR data"""
    storage = storage_root / "edr"
    
    def __init__(
        self,
        id_:str,  # CTX product id (pid)
        source_dir:str='',  # alternative root folder for EDR data
        with_volume:bool=True,  # does the storage path include the volume folder
        with_id_dir:bool=False,  # does the storage path include an extra pid folder
    ):
        store_attr(but='source_dir')
        self.storage = Path(source_dir) if source_dir else self.storage
        
    @property
    def pid(self):
        return self.id_
    
    @pid.setter
    def pid(self, value):
        self.id_ = value
        
    @property
    def meta(self):
        "get the metadata from the index table"
        s = edrindex.query("PRODUCT_ID == @self.pid").squeeze()
        s.index = s.index.str.lower()
        return s

    @property
    def volume(self):
        "get the PDS volume number for the current product id"
        return self.meta.volume_id.lower()
    
    @property
    def source_folder(self):
        if self.with_volume:
            base = self.storage / self.volume
        else:
            base = self.storage
        if not self.with_id_dir:
            return base
        else: 
            return base / self.pid

    @property
    def source_path(self):
        return self.source_folder / f"{self.pid}.IMG"

    @property
    def url(self):
        "Calculate URL from input dataframe row."
        url = baseurl / self.meta.volume_id.lower() / "data" / (self.pid + ".IMG")
        return url

    def download(self, overwrite=False):
        if self.source_path.exists() and not overwrite:
            print("File exists. Use `overwrite=True` to download fresh.")
            return
        self.source_folder.mkdir(parents=True, exist_ok=True)
        url_retrieve(self.url, self.source_path)


# %% ../notebooks/03_ctx.ipynb 12
class CTX:
    """Class to manage dealing with CTX data.
    
    HAS a CTXEDR attribute as defined above.
    """
    proc_dir = storage_root / "edr"

    def __init__(
        self, 
        id_:str,  # CTX product id
        source_dir:str='',  # where the raw EDR data is stored, if not coming from plpy
        proc_dir:str='',  # where to store processed, if not plpy
        with_volume:bool=False,  # store with extra volume subfolder?
        with_id_dir:bool=True  # store with extra product_id subfolder?
    ):
        store_attr(but="source_dir,proc_dir")
        self.proc_dir = Path(proc_dir) if proc_dir else self.proc_dir
        self.edr = CTXEDR(id_, source_dir, with_volume, with_id_dir)
        
        (self.cub_name, self.cal_name, self.destripe_name) = file_variations(
            self.edr.source_path.name, [".cub", ".cal.cub", ".dst.cal.cub"]
        )
        self.is_read = False
        self.is_calib_read = False

    @property
    def pid(self):
        return self.edr.pid

    @property
    def proc_folder(self):
        "the folder for all processed data. could be same as source_dir"
        return self.proc_dir / self.edr.source_folder.relative_to(self.edr.source_folder)

    @property
    def cub_path(self):
        return self.proc_folder / self.cub_name
    
    @property
    def cal_path(self):
        return self.proc_folder / self.cal_name
    
    @property
    def destripe_path(self):
        return self.proc_folder / self.destripe_name


    @catch_isis_error
    def isis_import(self):
        mroctx2isis(from_=self.edr.source_path, to=self.cub_path)

    @catch_isis_error
    def spice_init(self):
        spiceinit(from_=self.cub_path, web="yes")

    @catch_isis_error
    def calibrate(self):
        ctxcal(from_=self.cub_path, to=self.cal_path)
        self.is_calib_read = False

    @catch_isis_error
    def destripe(self):
        if self.do_destripe():
            ctxevenodd(from_=self.cal_path, to=self.destripe_path)
            self.destripe_path.rename(self.cal_path)

    @catch_isis_error
    def do_destripe(self):
        value = int(
            getkey(
                from_=self.cal_path,
                objname="isiscube",
                grpname="instrument",
                keyword="SpatialSumming",
            )
        )
        return False if value == 2 else True

    def calib_pipeline(self, overwrite=False):
        if self.cal_path.exists() and not overwrite:
            return
        pbar = tqdm("isis_import spice_init calibrate destripe".split())
        for name in pbar:
            pbar.set_description(name)
            getattr(self, name)()
        pbar.set_description("Done.")

    def read_edr(self):
        "`da` stands for dataarray, standard abbr. within xarray."
        if not self.edr.source_path.exists():
            raise FileNotFoundError("EDR not downloaded yet.")
        if not self.is_read:
            self.edr_da = rxr.open_rasterio(self.edr.source_path)
            self.is_read = True
        return self.edr_da

    def read_calibrated(self):
        "`da` stands for dataarray, standard abbr. within xarray."
        if not self.is_calib_read:
            self.cal_da = rxr.open_rasterio(self.cal_path)
            self.is_calibd_read = True
        return self.cal_da

    def plot_da(self, data=None):
        data = self.edr_da if data is None else data
        return data.isel(band=0, drop=True).hvplot(
            x="y", y="x", rasterize=True, cmap="gray", data_aspect=1
        )

    def plot_calibrated(self):
        return self.plot_da(self.read_calibrated())

    def __str__(self):
        s = f"PRODUCT_ID: {self.edr.pid}\n"
        s += f"URL: {self.edr.url}\n"
        s += f"Local: {self.edr.source_path}\n"
        try:
            s += f"Shape: {self.read_edr().shape}"
        except FileNotFoundError:
            s += f"Not downloaded yet."
        return s

    def __repr__(self):
        return self.__str__()

# %% ../notebooks/03_ctx.ipynb 35
@call_parse
def ctx_calib(
    id_:str,  # CTX product_id
    source:str='',  # path to where EDRs are stored if not from plpy
    proc_dir:str='',  # path to where processed data is to be stored
    overwrite:bool=False,  # overwrite processed data
):
    ctx = CTX(id_, source_dir=source, proc_dir=proc_dir)
    ctx.calib_pipeline(overwrite=overwrite)
    print("Produced\n", ctx.cal_path)

# %% ../notebooks/03_ctx.ipynb 37
class CTXEDRCollection:
    """Class to deal with a set of CTX products."""

    def __init__(self, product_ids):
        self.product_ids = product_ids

    def get_urls(self):
        """Get URLs for list of product_ids.

        Returns
        -------
        List[yarl.URL]
            List of URL objects with the respective PDS URL for download.
        """
        urls = []
        for p_id in self.product_ids:
            ctx = CTXEDR(p_id)
            urls.append(ctx.url)
        self.urls = urls
        return urls

    def download_collection(self):
        lazys = []
        for p_id in self.product_ids:
            ctx = CTXEDR(p_id)
            lazys.append(delayed(ctx.download)())
        print("Launching parallel download...")
        compute(*lazys)
        print("Done.")

    def calibrate_collection(self):
        lazys = []
        for p_id in self.product_ids:
            ctx = CTXEDR(p_id)
            lazys.append(delayed(ctx.calib_pipeline)())
        print("Launching parallel calibration...")
        compute(*lazys)
        print("Done.")

    def calib_exist_check(self):
        return [(p_id, CTXEDR(p_id).cal_name.exists()) for p_id in self.product_ids]



# AUTOGENERATED! DO NOT EDIT! File to edit: ../notebooks/05_cassini_uvis.ipynb.

# %% auto 0
__all__ = ['storage_root', 'DataManager', 'get_data_path', 'get_label_path', 'get_user_guide']

# %% ../notebooks/05_cassini_uvis.ipynb 3
import tomlkit
from yarl import URL

from .config import config
from .pds.opusapi import OPUS
from .utils import url_retrieve

# %% ../notebooks/05_cassini_uvis.ipynb 4
storage_root = config.storage_root / "missions/cassini/uvis"
storage_root

# %% ../notebooks/05_cassini_uvis.ipynb 5
class DataManager:
    def __init__(
        self, 
        pid: str,  # Product ID. If longer than PDS_ID, will be cut in attribute `pds_id`
        skip_download: bool=False  # skip trying to download
    ):
        self.pid = pid
        self.dict = None
        if not self.raw_data_path.exists() and not skip_download:
            self.download()

    def query(self, pds_id=None):
        pds_id = pds_id if pds_id is not None else self.pds_id
        opus = OPUS(silent=True)
        try:
            self.query_result = opus.query_image_id(pds_id)[0]
        except IndexError:
            raise FileNotFoundError("Project ID not found on PDS server.")
        self.opus_id = self.query_result[0]
        self.dict = self.query_result[1]

    @property
    def pds_id(self):
        return self.pid[:17]

    @property
    def folder(self):
        #         return storage_root / "/".join(self.raw_data_url.parts[4:7])
        return storage_root / self.pds_id

    @property
    def raw_data_url(self):
        if not self.dict:
            self.query()
        return URL(self.dict["couvis_raw"][0])

    @property
    def raw_label_url(self):
        if not self.dict:
            self.query()
        return URL(self.dict["couvis_raw"][1])

    @property
    def raw_data_path(self):
        return self.folder / (self.pds_id + ".DAT")

    @property
    def raw_label_path(self):
        return self.raw_data_path.with_suffix(".LBL")

    @property
    def calib_corr_path(self):
        return self.raw_data_path.with_name(self.raw_data_path.stem + "_CAL_3.DAT")

    @property
    def calib_label_path(self):
        return self.calib_corr_path.with_suffix(".LBL")

    @property
    def original_pid_file(self):
        return self.folder / "original_pid.txt"

    @property
    def results_file(self):
        return self.folder / "urls.toml"

    def download(self, overwrite=False):
        if self.raw_data_path.exists() and not overwrite:
            print("Local files exists. Use `overwrite=True` to download fresh.")
            return
        self.query()
        self.original_pid_file.mk_write(self.pid)
        self.results_file.mk_write(tomlkit.dumps(self.dict))
        self.raw_data_path.parent.mkdir(parents=True, exist_ok=True)
        for key in ["couvis_raw", "couvis_calib_corr"]:
            for url in self.dict[key]:
                url_retrieve(url, self.folder / URL(url).name)

    def __repr__(self):
        s = f"Product ID:\n{self.id}\n\n"
        for k, v in self.query_result[1].items():
            s += f"Key: {k},\nValue(s):\n{v}\n\n"
        return s

# %% ../notebooks/05_cassini_uvis.ipynb 19
def get_data_path(pid, skip_download=False):
    dm = DataManager(pid, skip_download=skip_download)
    return dm.raw_data_path if dm.raw_data_path.exists() else None


def get_label_path(pid):
    dm = DataManager(pid)
    return dm.raw_label_path

# %% ../notebooks/05_cassini_uvis.ipynb 22
def get_user_guide():
    url = URL(
        "https://pds-rings.seti.org/cassini/uvis/1-UVIS_Users_Guide_-2018-Jan%2015-For%20PDS-REV-2018-07-06.pdf"
    )
    local_path = storage_root / "uvis_user_guide.pdf"
    if not local_path.exists():
        url_retrieve(url, storage_root / "uvis_user_guide.pdf")
    return local_path



# AUTOGENERATED! DO NOT EDIT! File to edit: ../../notebooks/02b_pds.ctx_index.ipynb.

# %% auto 0
__all__ = ['CTXIndex']

# %% ../../notebooks/02b_pds.ctx_index.ipynb 3
from dataclasses import dataclass
from string import Template

import pandas as pd


@dataclass
class CTXIndex:
    """Class to determine the URL for the latest cumulative index.

    This is a 2 step process, where first the MRO release page is scraped
    for the latest CTX release, and then the latest release page is scraped
    for the latest volume.
    From that latest volume the latest index URL is constructed.
    """
    volumes_url: str = "https://pds-imaging.jpl.nasa.gov/volumes/mro.html"
    release_url_template: Template = Template(
        "https://pds-imaging.jpl.nasa.gov/volumes/mro/release${release}.html"
    )
    volume_url_template: Template = Template(
        "https://pds-imaging.jpl.nasa.gov/data/mro/mars_reconnaissance_orbiter/ctx/mrox_${volume}/"
    )
    scraped_tables: bool = False
    release_scraped: bool = False

    @property
    def web_tables_list(self):
        """Use the pandas scraper to read in the MRO data release table.

        The scraper returns several tables in a list and the last one
        lists all the CTX volumes.

        This could be replaced by cached properties.
        """
        if not self.scraped_tables:
            self._list_of_scraped_tables = pd.read_html(self.volumes_url)
            self.scraped_tables = True
        return self._list_of_scraped_tables

    @property
    def release_number(self):
        """Fishes out the release number.

        This is needed to construct the exact URL to the latest cumulative index file.
        """
        alist = self.web_tables_list
        return alist[-1].iloc[-1, 0].split()[-1]

    @property
    def release_url(self):
        "Constructs the release URL from the release number."
        return self.release_url_template.substitute(release=self.release_number)

    @property
    def latest_volume_url(self):
        """Scrape the Release URL for the latest volume URL in that.

        This is necessary because a release usually has more that one volume.
        """
        if not self.release_scraped:
            alist = pd.read_html(self.release_url)
            # get last row of 4th table
            row = alist[3].iloc[-1]
            number = None
            # first number that is NAN breaks the loop over last row of table
            for elem in row.values:
                try:
                    number = int(elem.split()[-1])
                except AttributeError:
                    break
            self.number = number
            self.release_scraped = True
        return self.volume_url_template.substitute(volume=self.number)

    @property
    def latest_index_label_url(self):
        "Construct the URL for the latest cumulative index."

        return self.latest_volume_url + "index/cumindex.lbl"



# AUTOGENERATED BY NBDEV! DO NOT EDIT!

__all__ = ["index", "modules", "custom_doc_links", "git_url"]

index = {"Config": "00_config.ipynb",
         "config": "00_config.ipynb",
         "logger": "04_hirise.ipynb",
         "nasa_date_format": "01_utils.ipynb",
         "nasa_dt_format": "01_utils.ipynb",
         "nasa_dt_format_with_ms": "01_utils.ipynb",
         "iso_date_format": "01_utils.ipynb",
         "iso_dt_format": "01_utils.ipynb",
         "iso_dt_format_with_ms": "01_utils.ipynb",
         "nasa_time_to_datetime": "01_utils.ipynb",
         "nasa_time_to_iso": "01_utils.ipynb",
         "iso_to_nasa_time": "01_utils.ipynb",
         "iso_to_nasa_datetime": "01_utils.ipynb",
         "replace_all_nasa_times": "01_utils.ipynb",
         "parse_http_date": "01_utils.ipynb",
         "get_remote_timestamp": "01_utils.ipynb",
         "check_url_exists": "01_utils.ipynb",
         "url_retrieve": "01_utils.ipynb",
         "have_internet": "01_utils.ipynb",
         "height_from_shadow": "01_utils.ipynb",
         "get_gdal_center_coords": "01_utils.ipynb",
         "file_variations": "01_utils.ipynb",
         "storage_root": "06_cassini_iss.ipynb",
         "dynamic_urls": "02a_pds.indexes.ipynb",
         "Index": "02a_pds.indexes.ipynb",
         "CTXIndex": "02b_pds.ctx_index.ipynb",
         "find_indexes": "02c_pds.apps.ipynb",
         "get_index": "02c_pds.apps.ipynb",
         "find_instruments": "02c_pds.apps.ipynb",
         "refresh_index": "02c_pds.apps.ipynb",
         "OPUSImageURL": "02d_pds.opusapi.ipynb",
         "OPUSObsID": "02d_pds.opusapi.ipynb",
         "OPUS": "02d_pds.opusapi.ipynb",
         "base_url": "10_spice.kernels.ipynb",
         "metadata_url": "02d_pds.opusapi.ipynb",
         "image_url": "02d_pds.opusapi.ipynb",
         "dic": "02d_pds.opusapi.ipynb",
         "LROCIndex": "02e_pds.lroc_index.ipynb",
         "IndexLabel": "02f_pds.utils.ipynb",
         "index_to_df": "02f_pds.utils.ipynb",
         "PVLColumn": "02f_pds.utils.ipynb",
         "decode_line": "02f_pds.utils.ipynb",
         "find_mixed_type_cols": "02f_pds.utils.ipynb",
         "fix_hirise_edrcumindex": "02f_pds.utils.ipynb",
         "baseurl": "04_hirise.ipynb",
         "edrindex": "03_ctx.ipynb",
         "catch_isis_error": "03_ctx.ipynb",
         "CTXEDR": "03_ctx.ipynb",
         "CTXEDRCollection": "03_ctx.ipynb",
         "rdrindex": "04_hirise.ipynb",
         "OBSID": "04_hirise.ipynb",
         "ProductPathfinder": "04_hirise.ipynb",
         "COLOR_PRODUCT": "04_hirise.ipynb",
         "RGB_NOMAP": "04_hirise.ipynb",
         "RGB_NOMAPCollection": "04_hirise.ipynb",
         "SOURCE_PRODUCT": "04_hirise.ipynb",
         "RED_PRODUCT": "04_hirise.ipynb",
         "IR_PRODUCT": "04_hirise.ipynb",
         "BG_PRODUCT": "04_hirise.ipynb",
         "DataManager": "05_cassini_uvis.ipynb",
         "get_data_path": "07_diviner.ipynb",
         "get_label_path": "05_cassini_uvis.ipynb",
         "get_user_guide": "05_cassini_uvis.ipynb",
         "opus_keys": "06_cassini_iss.ipynb",
         "ISSOpus": "06_cassini_iss.ipynb",
         "hostname": "07_diviner.ipynb",
         "calculate_image_azimuth": "08_geotools.ipynb",
         "get_north_shifted_point": "08_geotools.ipynb",
         "calculate_image_north_azimuth": "08_geotools.ipynb",
         "get_sun_angles": "08_geotools.ipynb",
         "debug_srs": "08_geotools.ipynb",
         "shift_to_center": "08_geotools.ipynb",
         "pixel_to_meter": "08_geotools.ipynb",
         "Point": "08_geotools.ipynb",
         "Window": "08_geotools.ipynb",
         "ImgData": "08_geotools.ipynb",
         "Error": "09_exceptions.ipynb",
         "SomethingNotSetError": "09_exceptions.ipynb",
         "ProjectionNotSetError": "09_exceptions.ipynb",
         "GeoTransformNotSetError": "09_exceptions.ipynb",
         "KMASpiceError": "09_exceptions.ipynb",
         "SPointNotSetError": "09_exceptions.ipynb",
         "ObserverNotSetError": "09_exceptions.ipynb",
         "kernel_storage": "10_spice.kernels.ipynb",
         "dataset_ids": "10_spice.kernels.ipynb",
         "df": "10_spice.kernels.ipynb",
         "df2": "10_spice.kernels.ipynb",
         "df2.columns": "10_spice.kernels.ipynb",
         "datasets": "10_spice.kernels.ipynb",
         "is_start_valid": "10_spice.kernels.ipynb",
         "is_stop_valid": "10_spice.kernels.ipynb",
         "Subsetter": "10_spice.kernels.ipynb",
         "get_metakernel_and_files": "10_spice.kernels.ipynb",
         "list_kernels_for_day": "10_spice.kernels.ipynb"}

modules = ["config.py",
           "utils.py",
           "pds/indexes.py",
           "pds/ctx_index.py",
           "pds/apps.py",
           "pds/opusapi.py",
           "pds/lroc_index.py",
           "pds/utils.py",
           "ctx.py",
           "hirise.py",
           "uvis.py",
           "ciss.py",
           "diviner.py",
           "geotools.py",
           "exceptions.py",
           "spice/kernels.py"]

doc_url = "https://michaelaye.github.io/nbplanetary/"

git_url = "https://github.com/michaelaye/nbplanetary/tree/master/"

def custom_doc_links(name): return None



# AUTOGENERATED! DO NOT EDIT! File to edit: ../notebooks/04_hirise.ipynb.

# %% auto 0
__all__ = ['logger', 'storage_root', 'baseurl', 'rdrindex', 'OBSID', 'ProductPathfinder', 'COLOR_PRODUCT', 'RGB_NOMAP',
           'RGB_NOMAPCollection', 'SOURCE_PRODUCT', 'RED_PRODUCT', 'IR_PRODUCT', 'BG_PRODUCT']

# %% ../notebooks/04_hirise.ipynb 3
import logging
import warnings
import webbrowser

import rasterio
import rioxarray as rxr
from dask import compute, delayed
from fastcore.utils import Path
from yarl import URL

import hvplot.xarray  # noqa
from .config import config
from .pds.apps import get_index
from .utils import check_url_exists, url_retrieve

warnings.filterwarnings("ignore", category=rasterio.errors.NotGeoreferencedWarning)

# %% ../notebooks/04_hirise.ipynb 4
logger = logging.getLogger(__name__)

# %% ../notebooks/04_hirise.ipynb 5
storage_root = config.storage_root / "missions/mro/hirise"
baseurl = URL("https://hirise-pds.lpl.arizona.edu/PDS")
rdrindex = get_index("mro.hirise", "rdr")

# %% ../notebooks/04_hirise.ipynb 6
class OBSID:
    """Manage HiRISE observation ids.

    For example PSP_003092_0985.
    `phase` is set to PSP for orbits < 11000, no setting required.
    """

    def __init__(
        self,
        obsid: str,  # e.g. PSP_003092_0985
    ):
        phase, orbit, targetcode = obsid.split("_")
        self._orbit = int(orbit)
        self._targetcode = targetcode

    @property
    def orbit(self):
        return str(self._orbit).zfill(6)

    @orbit.setter
    def orbit(self, value: int):  # e.g. 11000, < 1_000_000
        if value > 999999:
            raise ValueError("Orbit cannot be larger than 999999")
        elif len(value) != 6:
            raise ValueError("Orbit string must be 6 digits.")
        self._orbit = value

    @property
    def targetcode(self):
        return self._targetcode

    @targetcode.setter
    def targetcode(
        self,
        value: str,  # e.g. "0985", must be 4 digits
    ):
        if len(str(value)) != 4:
            raise ValueError("Targetcode must be exactly 4 characters.")
        self._targetcode = value

    @property
    def phase(self):
        return "PSP" if int(self.orbit) < 11000 else "ESP"

    @property
    def id(self):
        return f"{self.phase}_{self.orbit}_{self.targetcode}"

    def __repr__(self):
        return self.__str__()

    def __str__(self):
        return self.id

    @property
    def upper_orbit_folder(self):
        """
        get the upper folder name wa
        rdrhere the given orbit folder is residing on the
        hisync server, e.g. 'ORB_011900_011999'
        """
        lower = int(self.orbit) // 100 * 100
        lowerstr = str(lower).zfill(6)
        upperstr = str(lower + 99).zfill(6)
        return f"ORB_{lowerstr}_{upperstr}"

    @property
    def storage_path_stem(self):
        return f"{self.phase}/{self.upper_orbit_folder}/{self.id}"

# %% ../notebooks/04_hirise.ipynb 13
class ProductPathfinder:
    """Determine paths and URLs for HiRISE RDR products (also EXTRAS.)

    We use the PDS definition of PRODUCT_ID here, e.g. PSP_003092_0985_RED.

    Attributes `jp2_path` and `label_path` get you the official RDR mosaic product,
    with `kind` steering if you get the COLOR or the RED product.
    All other properties go to the RDR/EXTRAS folder.
    The "PDS" part of the path is handled in the OBSID class.
    """

    kinds = ["RED", "BG", "IR", "COLOR", "IRB", "MIRB", "MRGB", "RGB", "RGB.NOMAP"]

    @classmethod
    def from_path(cls, path):
        path = Path(path)
        return cls(path.stem)

    def __init__(
        self,
        initstr: str,  # PRODUCT_ID string, e.g. PSP_003092_0985_RED
        check_url: bool = True,  # for performance, the user might not want the url check
    ):
        tokens = initstr.split("_")
        self._obsid = OBSID("_".join(tokens[:3]))
        try:
            self.kind = tokens[3]
        except IndexError:
            self._kind = None
        self.check_url = check_url

    @property
    def obsid(self):
        return self._obsid

    @obsid.setter
    def obsid(self, value: str):  # e.g. "PSP_003092_0985"
        self._obsid = OBSID(value)

    @property
    def kind(self):
        return self._kind

    @kind.setter
    def kind(
        self,
        value: str,  # one of "RED", "BG", "IR", "COLOR", "IRB", "MIRB", "MRGB", "RGB", "RGB.NOMAP"
    ):
        if value not in self.kinds:
            raise ValueError(f"kind must be in {self.kinds}")
        self._kind = value

    @property
    def product_id(self):
        return f"{self.obsid}_{self.kind}"

    def __str__(self) -> str:  # PRODUCT_ID string, e.g. PSP_003092_0985_RED
        return self.product_id

    def __repr__(self):
        return self.__str__()

    @property
    def storage_stem(
        self,
    ) -> str:  # e.g. 'PSP/ORB_003000_003099/PSP_003092_0985/PSP_003092_0985_RED'
        return f"{self.obsid.storage_path_stem}/{self.product_id}"

    @property
    def label_fname(self) -> str:  # e.g. 'PSP_003092_0985_RED.LBL'
        return f"{self.product_id}.LBL"

    @property
    def label_path(self):
        return Path("RDR") / (self.storage_stem + ".LBL")

    @property
    def local_label_path(self):
        path = self.label_path
        return storage_root / (f"{path.parent.name}/{path.name}")

    def download_label(self, overwrite=False):
        """Download the label file."""
        self.local_label_path.parent.mkdir(exist_ok=True, parents=True)
        if self.local_label_path.exists and not overwrite:
            return
        else:
            url_retrieve(self.label_url, self.local_label_path)

    def _make_url(self, obj):
        path = getattr(self, f"{obj}_path")
        url = baseurl / str(path)
        if self.check_url:
            if not check_url_exists(url):
                warnings.warn(f"{url} does not exist on the server.")
        return url

    def __getattr__(self, item):
        tokens = item.split("_")
        try:
            if tokens[-1] == "url":
                return self._make_url("_".join(tokens[:-1]))
        except IndexError:
            raise AttributeError(f"No attribute named '{item}' found.")

    @property
    def jp2_fname(self):
        return self.product_id + ".JP2"

    @property
    def jp2_path(self):
        prefix = "RDR/"
        postfix = ""
        if self.kind not in ["RED", "COLOR"]:
            prefix += "EXTRAS/"
        if self.kind in ["IRB"]:
            postfix = ".NOMAP"
        return prefix + self.storage_stem + postfix + ".JP2"

    @property
    def nomap_jp2_path(self):
        if self.kind in ["RED", "IRB", "RGB"]:
            return f"EXTRAS/RDR/{self.storage_stem}.NOMAP.JP2"
        else:
            raise ValueError(f"No NOMAP exists for {self.kind}.")

    @property
    def quicklook_path(self):
        if self.kind in ["COLOR", "RED"]:
            return Path("EXTRAS/RDR/") / (self.storage_stem + ".QLOOK.JP2")
        else:
            raise ValueError(f"No quicklook exists for {self.kind} products.")

    @property
    def abrowse_path(self):
        if self.kind in ["COLOR", "MIRB", "MRGB", "RED"]:
            return Path("EXTRAS/RDR/") / (self.storage_stem + ".abrowse.jpg")
        else:
            raise ValueError(f"No abrowse exists for {self.kind}")

    @property
    def browse_path(self):
        inset = ""
        if self.kind in ["IRB", "RGB"]:
            inset = ".NOMAP"
        if self.kind not in ["COLOR", "MIRB", "MRGB", "RED", "IRB", "RGB"]:
            raise ValueError(f"No browse exists for {self.kind}")
        else:
            return Path("EXTRAS/RDR/") / (self.storage_stem + inset + ".browse.jpg")

    @property
    def thumbnail_path(self):
        if self.kind in ["BG", "IR"]:
            raise ValueError(f"No thumbnail exists for {self.kind}")
        inset = ""
        if self.kind in ["IRB", "RGB"]:
            inset = ".NOMAP"
        return Path("EXTRAS/RDR/") / (self.storage_stem + inset + ".thumb.jpg")

    @property
    def nomap_thumbnail_path(self):
        if self.kind in ["RED", "IRB", "RGB"]:
            return Path("EXTRAS/RDR") / (self.storage_stem + ".NOMAP.thumb.jpg")
        else:
            raise ValueError(f"No NOMAP thumbnail exists for {self.kind} images.")

    @property
    def nomap_browse_path(self):
        if self.kind in ["RED", "IRB", "RGB"]:
            return Path("EXTRAS/RDR") / (self.storage_stem + ".NOMAP.browse.jpg")
        else:
            raise ValueError(f"No NOMAP browse exists for {self.kind} images.")

    @property
    def edr_storage_stem(self):
        return Path("EDR") / self.storage_stem

    @property
    def homepage(self) -> str:  # URL to the product's homepage
        return f"https://uahirise.org/{self.obsid}"

    def go_to_homepage(self):
        webbrowser.open(self.homepage)

# %% ../notebooks/04_hirise.ipynb 33
class COLOR_PRODUCT:
    def __init__(self, obsid):
        self.obsid = obsid
        # this should be reset by the subclass
        self.pathfinder = ProductPathfinder(self.obsid + "_COLOR")

    @property
    def product_id(self):
        return self.pathfinder.product_id

    @property
    def meta(self):
        color_id = self.obsid + "_COLOR"
        s = rdrindex.query("PRODUCT_ID == @color_id").squeeze()
        s.index = s.index.str.lower()
        return s

    @property
    def url(self):
        # set `name` in the subclass
        return URL(getattr(self.pathfinder, self.name + "_url"))

    @property
    def remote_path(self):
        # set `name` in subclass
        return Path(getattr(self.pathfinder, self.name + "_path"))

    @property
    def local_path(self):
        path = self.remote_path
        return storage_root / (f"{path.parent.name}/{path.name}")

    def download(self, overwrite=False):
        self.local_path.parent.mkdir(parents=True, exist_ok=True)
        if self.local_path.exists() and not overwrite:
            print("File exists. Use `overwrite=True` to download fresh.")
            return
        url_retrieve(self.url, self.local_path)

    def read(self):
        self.da = rxr.open_rasterio(self.local_path, chunks=(1, 2024, 2024))
        return self.da

    def show(self, **kwargs):
        self.read()
        return self.plot_da(**kwargs)

    def plot_da(self, xslice=None, yslice=None):
        if xslice is not None or yslice is not None:
            data = self.da.isel(x=xslice, y=yslice)
        else:
            data = self.da

        return data.hvplot.image(
            x="x",
            y="y",
            rasterize=True,
            widget_location="top_left",
            cmap="gray",
            frame_height=800,
            frame_width=800,
            flip_yaxis=True,
        )

# %% ../notebooks/04_hirise.ipynb 34
class RGB_NOMAP(COLOR_PRODUCT):
    def __init__(self, obsid):
        super().__init__(obsid)
        self.name = "nomap_jp2"
        self.pathfinder = ProductPathfinder(obsid + "_RGB")

# %% ../notebooks/04_hirise.ipynb 45
class RGB_NOMAPCollection:
    """Class to deal with a set of RGB_NOMAP products."""

    def __init__(self, obsids):
        self.obsids = obsids

    def get_urls(self):
        """Get URLs for list of obsids.

        Returns
        -------
        List[yarl.URL]
            List of URL objects with the respective PDS URL for download.
        """
        urls = []
        for obsid in self.obsids:
            rgb = RGB_NOMAP(obsid)
            urls.append(rgb.url)
        self.urls = urls
        return urls

    @property
    def local_paths(self):
        paths = []
        for obsid in self.obsids:
            rgb = RGB_NOMAP(obsid)
            paths.append(rgb.local_path)
        return paths

    def download_collection(self):
        lazys = []
        for obsid in self.obsids:
            rgb = RGB_NOMAP(obsid)
            lazys.append(delayed(rgb.download)())
        print("Launching parallel download...")
        compute(*lazys)
        print("Done.")

# %% ../notebooks/04_hirise.ipynb 46
class SOURCE_PRODUCT:
    """Manage SOURCE_PRODUCT id.

    Example
    -------
    'PSP_003092_0985_RED4_0'
    """

    red_ccds = ["RED" + str(i) for i in range(10)]
    ir_ccds = ["IR10", "IR11"]
    bg_ccds = ["BG12", "BG13"]
    ccds = red_ccds + ir_ccds + bg_ccds

    def __init__(self, spid, saveroot=None, check_url=True):
        tokens = spid.split("_")
        obsid = "_".join(tokens[:3])
        ccd = tokens[3]
        color, ccdno = self._parse_ccd(ccd)
        self.pid = ProductPathfinder("_".join([obsid, color]))
        self.ccd = ccd
        self.channel = tokens[4]
        self.saveroot = storage_root if saveroot is None else saveroot

    def __getattr__(self, value):
        return getattr(self.pid, value)

    def _parse_ccd(self, value):
        sep = 2 if value[:2] in ProductPathfinder.kinds else 3
        return value[:sep], value[sep:]

    @property
    def channel(self):
        return self._channel

    @channel.setter
    def channel(self, value):
        if int(value) not in [0, 1]:
            raise ValueError("channel must be in [0, 1]")
        self._channel = value

    @property
    def ccd(self):
        return self._ccd

    @ccd.setter
    def ccd(self, value):
        if value not in self.ccds:
            raise ValueError("CCD value must be in {}.".format(self.ccds))
        self._ccd = value
        if self.pid is not None:
            self.pid.color = self.color

    @property
    def color(self):
        return self._parse_ccd(self.ccd)[0]

    @property
    def ccdno(self):
        offset = len(self.color)
        return self.ccd[offset:]

    def __str__(self):
        return "{}: {}{}_{}".format(
            self.__class__.__name__, self.pid, self.ccdno, self.channel
        )

    def __repr__(self):
        return self.__str__()

    @property
    def spid(self):
        return f"{self.pid}{self.ccdno}_{self.channel}"

    @property
    def fname(self):
        return self.spid + ".IMG"

    @property
    def remote_path(self):
        return Path(self.pid.edr_storage_stem).parent / self.fname

    @property
    def url(self):
        u = baseurl / str(self.remote_path)
        if self.check_url:
            if not check_url_exists(u):
                warnings.warn(f"{u} does not exist on the server.")
        return u

    @property
    def local_path(self):
        savepath = self.saveroot / str(self.obsid) / self.fname
        return savepath

    @property
    def local_cube(self):
        return self.local_path.with_suffix(".cub")

    @property
    def stitched_cube_name(self):
        return f"{self.pid.obsid.id}_{self.ccd}.cub"

    @property
    def stitched_cube_path(self):
        return self.local_cube.with_name(self.stitched_cube_name)

    def download(self, overwrite=False):
        self.local_path.parent.mkdir(parents=True, exist_ok=True)
        if self.local_path.exists() and not overwrite:
            print("File exists. Use `overwrite=True` to download fresh.")
            return
        url_retrieve(self.url, self.local_path)

# %% ../notebooks/04_hirise.ipynb 61
class RED_PRODUCT(SOURCE_PRODUCT):
    "This exists to support creating a RED_PRODUCT_ID from parts of a SOURCE_PRODUCT id."

    def __init__(self, obsid, ccdno, channel, **kwargs):
        self.ccds = self.red_ccds
        super().__init__(f"{obsid}_RED{ccdno}_{channel}", **kwargs)

# %% ../notebooks/04_hirise.ipynb 66
class IR_PRODUCT(SOURCE_PRODUCT):
    def __init__(self, obsid, ccdno, channel):
        self.ccds = self.ir_ccds
        super().__init__(f"{obsid}_IR{ccdno}_{channel}", **kwargs)


class BG_PRODUCT(SOURCE_PRODUCT):
    def __init__(self, obsid, ccdno, channel):
        self.ccds = self.ir_ccds
        super().__init__(f"{obsid}_BG{ccdno}_{channel}", **kwargs)



# AUTOGENERATED! DO NOT EDIT! File to edit: ../notebooks/01_utils.ipynb.

# %% auto 0
__all__ = ['logger', 'nasa_date_format', 'nasa_dt_format', 'nasa_dt_format_with_ms', 'iso_date_format', 'iso_dt_format',
           'iso_dt_format_with_ms', 'nasa_time_to_datetime', 'nasa_time_to_iso', 'iso_to_nasa_time',
           'iso_to_nasa_datetime', 'replace_all_nasa_times', 'parse_http_date', 'get_remote_timestamp',
           'check_url_exists', 'url_retrieve', 'have_internet', 'height_from_shadow', 'get_gdal_center_coords',
           'file_variations']

# %% ../notebooks/01_utils.ipynb 3
import datetime as dt
import email.utils as eut
import http.client as httplib
import logging
from math import radians, tan
from pathlib import Path
from typing import Tuple, Union
from urllib.request import urlopen

import requests
from astropy.time import Time as ASTROTIME
from tqdm.auto import tqdm

import pandas as pd

try:
    from osgeo import gdal
except ImportError:
    GDAL_INSTALLED = False
else:
    GDAL_INSTALLED = True

# %% ../notebooks/01_utils.ipynb 4
logger = logging.getLogger(__name__)

if not GDAL_INSTALLED:
    logger.warning(
            "No GDAL found. Some planetary.utils functions not working, but okay."
        )


# %% ../notebooks/01_utils.ipynb 6
nasa_date_format = "%Y-%j"
nasa_dt_format = nasa_date_format + "T%H:%M:%S"
nasa_dt_format_with_ms = nasa_dt_format + ".%f"
iso_date_format = "%Y-%m-%d"
iso_dt_format = iso_date_format + "T%H:%M:%S"
iso_dt_format_with_ms = iso_dt_format + ".%f"

# %% ../notebooks/01_utils.ipynb 8
def _nasa_date_to_datetime(
    datestr: str,  # Date string of the form Y-j
) -> dt.datetime:
    "Convert date string to datetime."
    return dt.datetime.strptime(datestr, nasa_date_format)


def _nasa_datetime_to_datetime(
    datetimestr: str,  # datetime string of the form Y-jTH:M:S
) -> dt.datetime:
    "Convert datetime up to seconds to datetime."
    return dt.datetime.strptime(datetimestr, nasa_dt_format)


def _nasa_datetimems_to_datetime(
    datetimestr: str,  # datetime string of the form Y-jTH:M:S.xxx
) -> dt.datetime:
    "Convert date with millisec to datetime."
    return dt.datetime.strptime(datetimestr, nasa_dt_format_with_ms)

# %% ../notebooks/01_utils.ipynb 9
def nasa_time_to_datetime(
    inputstr,  # inputstr of format YYYY-jjj, YYYY-jjjTHH:MM:SS or YYYY-jjjTHH:MM:SS.ffffff
) -> dt.datetime:
    "User function to convert all kinds of NASA PDS datestrings with day_of_year into datetimes."
    try:
        return _nasa_datetime_to_datetime(inputstr)
    except ValueError:
        try:
            return _nasa_date_to_datetime(inputstr)
        except ValueError:
            return _nasa_datetimems_to_datetime(inputstr)

# %% ../notebooks/01_utils.ipynb 15
def nasa_time_to_iso(
    inputstr: str,
    with_hours: bool = False,  # Switch if return is wanted with hours (i.e. isoformat)
) -> str:  # Datestring in ISO-format.
    """Convert the day-number based NASA datetime format to ISO"""
    has_hours = False
    # check if input has hours
    try:
        res = _nasa_date_to_datetime(inputstr)
    except ValueError:
        has_hours = True
    time = nasa_time_to_datetime(inputstr)
    if has_hours or with_hours is True:
        return time.isoformat()
    else:
        return time.strftime(iso_date_format)

# %% ../notebooks/01_utils.ipynb 22
def iso_to_nasa_time(
    inputstr: str,  # Date string of the form Y-m-d
) -> str:  # Datestring in NASA standard yyyy-jjj
    "Convert iso date to day-number based NASA date."
    try:
        date = dt.datetime.strptime(inputstr, iso_date_format)
    except ValueError:
        try:
            date = dt.datetime.strptime(inputstr, iso_dt_format)
        except ValueError:
            date = dt.datetime.strptime(inputstr, iso_dt_format_with_ms)
            return date.strftime(nasa_dt_format_with_ms)
        else:
            return date.strftime(nasa_dt_format)
    else:
        return date.strftime(nasa_date_format)

# %% ../notebooks/01_utils.ipynb 23
def iso_to_nasa_datetime(
    dtimestr: str,  # Datetime string of the form yyyy-mm-ddTHH-MM-SS
):  # Datestring in NASA standard yyyy-jjjTHH-MM-SS
    "Convert iso datetime to day-number based NASA datetime."
    try:
        dtimestr.split(".")[1]
    except IndexError:
        source_format = iso_dt_format
        target_format = nasa_dt_format
    else:
        source_format = iso_dt_format_with_ms
        target_format = nasa_dt_format_with_ms
    date = dt.datetime.strptime(dtimestr, source_format)
    return date.strftime(target_format)

# %% ../notebooks/01_utils.ipynb 28
def replace_all_nasa_times(
    df: pd.DataFrame,  # DataFrame with NASA time columns
):
    """Find all NASA times in dataframe and replace with ISO.

    Changes will be implemented on incoming dataframe!

    This will be done for all columns with the word TIME in the column name.
    """
    for col in [col for col in df.columns if "TIME" in col]:
        if "T" in df[col].iloc[0]:
            df[col] = pd.to_datetime(df[col].map(nasa_time_to_iso))

# %% ../notebooks/01_utils.ipynb 30
def parse_http_date(
    text: str,  # datestring from urllib.request
) -> dt.datetime:  # dt.datetime object from given datetime string
    "Parse date string retrieved via urllib.request."
    return dt.datetime(*eut.parsedate(text)[:6])


def get_remote_timestamp(
    url: str,  # URL to check timestamp for
) -> dt.datetime:
    """Get the timestamp of a remote file.

    Useful for checking if there's an updated file available.
    """
    with urlopen(url, timeout=10) as conn:
        t = parse_http_date(conn.headers["last-modified"])
    return t


def check_url_exists(url):
    response = requests.head(url)
    if response.status_code < 400:
        return True
    else:
        return False


def url_retrieve(
    url: str,  # The URL to download
    outfile: str,  # The path where to store the downloaded file.
    # The size of the chunk for the request.iter_content call. Default: 128
    chunk_size: int = 128,
):
    """Improved urlretrieve with progressbar, timeout and chunker.

    This downloader has built-in progress bar using tqdm and using the `requests`
    package it improves standard `urllib` behavior by adding time-out capability.

    I tested different chunk_sizes and most of the time 128 was actually fastest, YMMV.

    Inspired by https://stackoverflow.com/a/61575758/680232
    """
    R = requests.get(url, stream=True, allow_redirects=True)
    if R.status_code != 200:
        raise ConnectionError(f"Could not download {url}\nError code: {R.status_code}")
    with tqdm.wrapattr(
        open(outfile, "wb"),
        "write",
        miniters=1,
        total=int(R.headers.get("content-length", 0)),
        desc=str(Path(outfile).name) + "\n",
    ) as fd:
        for chunk in R.iter_content(chunk_size=chunk_size):
            fd.write(chunk)


def have_internet():
    """Fastest way to check for active internet connection.

    From https://stackoverflow.com/a/29854274/680232
    """
    conn = httplib.HTTPConnection("www.google.com", timeout=5)
    try:
        conn.request("HEAD", "/")
        conn.close()
        return True
    except:
        conn.close()
        return False

# %% ../notebooks/01_utils.ipynb 32
def height_from_shadow(
    shadow_in_pixels: float,  # Measured length of shadow in pixels
    sun_elev: float,  # Ange of sun over horizon in degrees
) -> float:  # Height [meter]
    """Calculate height of an object from its shadow length.

    Note, that your image might have been binned.
    You need to correct `shadow_in_pixels` for that.
    """
    return tan(radians(sun_elev)) * shadow_in_pixels


def get_gdal_center_coords(
    imgpath: Union[str, Path],  # Path to raster image that is readable by GDLA
) -> Tuple[int, int]:  # center row/col coordinates.
    """Get center rows/cols pixel coordinate for GDAL-readable dataset.

    Check CLI `gdalinfo --formats` to see all formats that GDAL can open.
    """
    if not GDAL_INSTALLED:
        logger.error("GDAL not installed. Returning")
        return
    ds = gdal.Open(str(imgpath))
    xmean = ds.RasterXSize // 2
    ymean = ds.RasterYSize // 2
    return xmean, ymean


def file_variations(
    filename: Union[str, Path],  # The original filename to use as a base.
    extensions: list,
) -> list:  # list of Paths
    """Create a variation of file names.

    Generate a list of variations on a filename by replacing the extension with
    the provided list.

    Adapted from T. Olsens `file_variations of the pysis module for using pathlib.
    """
    return [Path(filename).with_suffix(extension) for extension in extensions]



# AUTOGENERATED! DO NOT EDIT! File to edit: ../../notebooks/10_spice.kernels.ipynb.

# %% auto 0
__all__ = ['kernel_storage', 'dataset_ids', 'df', 'df2', 'datasets', 'base_url', 'is_start_valid', 'is_stop_valid', 'Subsetter',
           'get_metakernel_and_files', 'list_kernels_for_day']

# %% ../../notebooks/10_spice.kernels.ipynb 3
import zipfile
from datetime import timedelta
from io import BytesIO
from pathlib import Path
import warnings
import requests

from astropy.time import Time
from dask.distributed import Client
from fastcore.utils import store_attr
from fastcore.test import test_fail
from tqdm.auto import tqdm
from yarl import URL
import spiceypy as spice

import pandas as pd
from ..config import config
from ..utils import nasa_time_to_iso, url_retrieve

# %% ../../notebooks/10_spice.kernels.ipynb 5
kernel_storage = config.storage_root / "spice_kernels"
kernel_storage.mkdir(exist_ok=True, parents=True)

# %% ../../notebooks/10_spice.kernels.ipynb 7
dataset_ids = {
    "cassini": "co-s_j_e_v-spice-6-v1.0/cosp_1000",
    "clementine": "clem1-l-spice-6-v1.0/clsp_1000",
    "dawn": "dawn-m_a-spice-6-v1.0/dawnsp_1000",
    "di": "di-c-spice-6-v1.0/disp_1000",
    "ds1": "ds1-a_c-spice-6-v1.0/ds1sp_1000",
    "epoxi": "dif-c_e_x-spice-6-v1.0/epxsp_1000",
    "em16": "em16/em16_spice",
    "grail": "grail-l-spice-6-v1.0/grlsp_1000",
    "hayabusa": "hay-a-spice-6-v1.0/haysp_1000",
    "insight": "insight/insight_spice",
    "juno": "jno-j_e_ss-spice-6-v1.0/jnosp_1000",
    "ladee": "ladee/ladee_spice",
    "lro": "lro-l-spice-6-v1.0/lrosp_1000",
    "maven": "maven/maven_spice",
    "opportunity": "mer1-m-spice-6-v1.0/mer1sp_1000",
    "spirit": "mer2-m-spice-6-v1.0/mer2sp_1000",
    "messenger": "mess-e_v_h-spice-6-v1.0/messsp_1000",
    "mars2020": "mars2020/mars2020_spice",
    "mex": "mex-e_m-spice-6-v2.0/mexsp_2000",
    "mgs": "mgs-m-spice-6-v1.0/mgsp_1000",
    "ody": "ody-m-spice-6-v1.0/odsp_1000",
    "mro": "mro-m-spice-6-v1.0/mrosp_1000",
    "msl": "msl-m-spice-6-v1.0/mslsp_1000",
    "near": "near-a-spice-6-v1.0/nearsp_1000",
    "nh": "nh-j_p_ss-spice-6-v1.0/nhsp_1000",
    "orex": "orex/orex_spice",
    "rosetta": "ro_rl-e_m_a_c-spice-6-v1.0/rossp_1000",
    "stardust": "sdu-c-spice-6-v1.0/sdsp_1000",
    "venus_climate_orbiter": "vco-v-spice-6-v1.0/vcosp_1000",
    "vex": "vex-e_v-spice-6-v2.0/vexsp_2000",
    "vo": "vo1_vo2-m-spice-6-v1.0/vosp_1000",
}

df = pd.DataFrame({"shorthand": dataset_ids.keys(), "path": dataset_ids.values()})

df2 = pd.read_html("https://naif.jpl.nasa.gov/naif/data_archived.html")[6]
df2.columns = df2.iloc[0]
df2 = df2.drop(0).reset_index(drop=True)
df2 = df2.drop(["Archive Readme", "Archive Link", "Subset Link"], axis=1)
df = df.join(df2)
datasets = df.set_index("shorthand")

# %% ../../notebooks/10_spice.kernels.ipynb 10
def is_start_valid(
    mission: str,  # mission shorthand label of datasets dataframe
    start: Time,  # start time in astropy.Time format
):
    return Time(datasets.at[mission, "Start Time"]) <= start


def is_stop_valid(
    mission: str,  # mission shorthand label of datasets dataframe
    stop: Time,  # stop time in astropy.Time format
):
    return Time(datasets.at[mission, "Stop Time"]) >= stop

# %% ../../notebooks/10_spice.kernels.ipynb 13
base_url = URL("https://naif.jpl.nasa.gov/cgi-bin/subsetds.pl")

# %% ../../notebooks/10_spice.kernels.ipynb 15
class Subsetter:
    """Class to manage retrieving subset SPICE kernel lists

    Attributes
    ----------
    kernel_names: List of
    """

    def __init__(
        self,
        mission: str,  # mission shorthand in datasets dataframe
        start: str,  # start time in either ISO or yyyy-jjj format
        stop=None,  # stop time in either ISO or yyyy-jjj format
        save_location=None,  # overwrite default storing in planetarpy archive
    ):
        store_attr()
        payload = self._get_payload()
        r = requests.get(base_url, params=payload, stream=True)
        if r.ok:
            z = zipfile.ZipFile(BytesIO(r.content))
        else:
            raise IOError("SPICE Server request returned status code: {r.status_code}")
        self.z = z
        # these files only exist "virtually" in the zip object, but are needed to
        # extract them:
        self.urls_file = [n for n in z.namelist() if n.startswith("urls_")][0]
        self.metakernel_file = [n for n in z.namelist() if n.lower().endswith(".tm")][0]
        with self.z.open(self.urls_file) as f:
            self.kernel_urls = f.read().decode().split()

    @property
    def start(self):
        return self._start

    @start.setter
    def start(self, value):
        try:
            self._start = Time(value)
        except ValueError:
            self._start = Time(nasa_time_to_iso(value))

    @property
    def stop(self):
        return self._stop

    @stop.setter
    def stop(self, value):
        if not value:
            self._stop = self.start + timedelta(days=1)
        else:
            try:
                self._stop = Time(value)
            except ValueError:
                self._stop = Time(nasa_time_to_iso(value))

    def _get_payload(self):
        """Put payload together while checking for working time format.

        If Time(self.start) doesn't work, then we assume that the date must be in the
        Time-unsupported yyyy-jjj format, which can be converted by `nasa_time_to_iso`
        from `planetarypy.utils`.
        """
        if not (
            is_start_valid(self.mission, self.start)
            and is_stop_valid(self.mission, self.stop)
        ):
            raise ValueError(
                "One of start/stop is outside the supported date-range. See `datasets`."
            )
        payload = {
            "dataset": dataset_ids[self.mission],
            "start": self.start.iso,
            "stop": self.stop.iso,
            "action": "Subset",
        }
        return payload

    @property
    def kernel_names(self):
        "Return list of names of kernels for the given time range."
        return [
            str(Path(URL(url).parent.name) / URL(url).name) for url in self.kernel_urls
        ]

    def get_local_path(
        self,
        url,  # kernel url to determine local storage path
    ) -> Path:  # full local path where kernel in URL will be stored
        """Calculate local storage path from Kernel URL, using `save_location` if given.

        If self.save_location is None, the `planetarypy` archive is being used.
        """
        u = URL(url)
        basepath = (
            kernel_storage / self.mission
            if not self.save_location
            else self.save_location
        )
        return basepath / u.parent.name / u.name

    def _non_blocking_download(
        self, 
        overwrite: bool = False
    ):
        with Client() as client:
            futures = []
            for url in tqdm(self.kernel_urls, desc="Kernels downloaded"):
                local_path = self.get_local_path(url)
                if local_path.exists() and not overwrite:
                    print(
                        local_path.parent.name,
                        local_path.name,
                        "locally available."
                    )
                    continue
                local_path.parent.mkdir(exist_ok=True, parents=True)
                futures.append(client.submit(url_retrieve, url, local_path))
            return [f.result() for f in futures]
            
    def download_kernels(
        self,
        overwrite: bool = False,  # switch to control if kernels should be downloaded over existing ones
        non_blocking: bool=False,
    ):
        if non_blocking:
            return self._non_blocking_download(overwrite)
        # sequential download
        for url in tqdm(self.kernel_urls, desc="Kernels downloaded"):
            local_path = self.get_local_path(url)
            if local_path.exists() and not overwrite:
                print(local_path.parent.name, local_path.name, "locally available.")
                continue
            local_path.parent.mkdir(exist_ok=True, parents=True)
            url_retrieve(url, local_path)

    def get_metakernel(self) -> Path:  # return path to metakernel file
        """Get metakernel file from NAIF and adapt path to match local storage.

        Use `save_location` if given, otherwise `planetarypy` archive.
        """
        basepath = (
            kernel_storage / self.mission
            if not self.save_location
            else self.save_location
        )
        savepath = basepath / self.metakernel_file
        with open(savepath, "w") as outfile, self.z.open(
            self.metakernel_file
        ) as infile:
            for line in infile:
                linestr = line.decode()
                if "'./data'" in linestr:
                    linestr = linestr.replace("'./data'", f"'{savepath.parent}'")
                outfile.write(linestr)
        return savepath

# %% ../../notebooks/10_spice.kernels.ipynb 43
def get_metakernel_and_files(
    mission: str,  # mission shorthand from datasets dataframe
    start: str,  # start time as iso-string, or yyyy-jjj
    stop: str,  # stop time as iso-string or yyyy-jjj
    save_location: str = None,  # override storage into planetarypy archive
) -> Path:  # pathlib.Path to metakernel file with corrected data path.
    "For a given mission and start/stop times, download the kernels and get metakernel path"
    subset = Subsetter(mission, start, stop, save_location)
    subset.download_kernels(non_blocking=True)
    return subset.get_metakernel()

# %% ../../notebooks/10_spice.kernels.ipynb 45
def list_kernels_for_day(
    mission: str,  # mission shorthand from datasets dataframe
    start: str,  # start time as iso-string, or yyyy-jjj
    stop: str = "",  # stop time as iso-string or yyyy-jjj
) -> list:  # list of kernel names
    subset = Subsetter(mission, start, stop)
    return subset.kernel_names
